# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

import numpy as np

from skbio.diversity._util import (
    _validate_counts_vector,
    _validate_taxa_and_tree,
    _vectorize_counts_and_tree,
    _check_taxa_alias,
)


def _setup_pd(counts, taxa, tree, validate, rooted, single_sample):
    if validate:
        if single_sample:
            # only validate count if operating in single sample mode, they
            # will have already been validated otherwise
            counts = _validate_counts_vector(counts)
            _validate_taxa_and_tree(counts, taxa, tree, rooted)
        else:
            _validate_taxa_and_tree(counts[0], taxa, tree, rooted)

    counts_by_node, _, branch_lengths = _vectorize_counts_and_tree(counts, taxa, tree)

    return counts_by_node, branch_lengths


def _faith_pd(counts_by_node, branch_lengths):
    """Calculate Faith's phylogenetic diversity (Faith's PD) metric.

    Parameters
    ----------
    counts_by_node : ndarray of shape (n_samples, n_nodes)
        Total counts/abundances of taxa descending from individual nodes of the tree.
    branch_lengths : ndarray of shape (n_nodes,)
        Branch lengths of corresponding nodes of the tree.

    Returns
    -------
    float
        Faith's phylogenetic diversity (PD).

    """
    return (branch_lengths * (counts_by_node > 0)).sum()


def faith_pd(counts, taxa=None, tree=None, validate=True, otu_ids=None):
    r"""Calculate Faith's phylogenetic diversity (Faith's PD) metric.

    The Faith's PD metric is defined as:

    .. math::

       PD = \sum_{b \in T \sqcup R} l(b)

    where :math:`T` is a minimum set of branches (:math:`b`) in a rooted tree
    that connect all taxa in a community. :math:`R` is a set of branches from
    the lowest common ancestor (LCA) of the taxa to the root of the tree.
    :math:`PD` is the sum of lengths (:math:`l`) of branches in both sets.

    Parameters
    ----------
    counts : 1-D array_like, int
        Vectors of counts/abundances of taxa for one sample.
    taxa : list, np.array
        Vector of taxon IDs corresponding to tip names in ``tree``. Must be the
        same length as ``counts``. Required.
    tree : skbio.TreeNode
        Tree relating taxa. The set of tip names in the tree can be a superset
        of ``taxa``, but not a subset. Required.
    validate : bool, optional
        If ``False``, validation of the input won't be performed. This step can
        be slow, so if validation is run elsewhere it can be disabled here.
        However, invalid input data can lead to invalid results or error
        messages that are hard to interpret, so this step should not be
        bypassed if you're not certain that your input data are valid. See
        :mod:`skbio.diversity` for the description of what validation entails
        so you can determine if you can safely disable validation.
    otu_ids : list, np.array
        Alias of ``taxa`` for backward compatibility. Deprecated and to be
        removed in a future release.

    Returns
    -------
    float
        Faith's phylogenetic diversity (PD).

    Raises
    ------
    ValueError, MissingNodeError, DuplicateNodeError
        If validation fails. Exact error will depend on what was invalid.

    See Also
    --------
    phydiv
    skbio.diversity
    skbio.diversity.alpha_diversity

    Notes
    -----
    Faith's phylogenetic diversity, often referred to as PD, was originally
    described in [1]_. It is the total phylogenetic branch length spanning
    all taxa of a community.

    It was clarified that the calculation should extend to the root of the
    tree [2]_, such that a single-taxon community will not have PD = 0. The
    root should be ancestral to all taxa being considered in the study, but
    does not need to be the origin of life. One should choose the root
    according to the scope of the study.

    Unrooted and abundance-weighted variants of PD are implemented in
    ``phydiv``.

    Several other metrics, such as evolutionary history (EH) [3]_ and
    functional diversity (FD) [4]_, are equivalent to PD in calculation.

    If computing Faith's PD for multiple samples, using
    ``skbio.diversity.alpha_diversity`` will be much faster than calling this
    function individually on each sample.

    This implementation of Faith's PD is based on the array-based
    implementation of UniFrac described in [5]_.

    This implementation differs from that in PyCogent (and therefore QIIME
    versions less than 2.0.0) by imposing a few additional restrictions on the
    inputs. First, the input tree must be rooted. In PyCogent, if an unrooted
    tree was provided that had a single trifurcating node (a newick convention
    for unrooted trees) that node was considered the root of the tree. Next,
    all taxa must be tips in the tree. PyCogent would silently ignore taxa that
    were not present the tree. To reproduce Faith PD results from PyCogent with
    scikit-bio, ensure that your PyCogent Faith PD calculations are performed
    on a rooted tree and that all taxa are present in the tree.

    References
    ----------
    .. [1] Faith, D. P. Conservation evaluation and phylogenetic diversity.
       Biol. Conserv. (1992).

    .. [2] Faith, D. P., & Baker, A. M. (2006). Phylogenetic diversity (PD)
       and biodiversity conservation: some bioinformatics challenges.
       Evolutionary bioinformatics, 2, 117693430600200007.

    .. [3] Nee, S., & May, R. M. (1997). Extinction and the loss of
       evolutionary history. Science, 278(5338), 692-694.

    .. [4] Petchey OL, Gaston KJ. Functional diversity (FD), species richness
       and community composition. Ecology letters. 2002 May;5(3):402-11.

    .. [5] Hamady M, Lozupone C, Knight R. Fast UniFrac: facilitating high-
       throughput phylogenetic analyses of microbial communities including
       analysis of pyrosequencing and PhyloChip data. ISME J. 4(1):17-27
       (2010).

    Examples
    --------
    Assume we have the following abundance data for a sample ``u``, represented
    as a counts vector. These counts represent the number of times specific
    taxa were observed in the sample.

    >>> u_counts = [1, 0, 0, 4, 1, 2, 3, 0]

    Because Faith PD is a phylogenetic diversity metric, we need to know which
    taxon each count corresponds to, which we'll provide as ``taxa``.

    >>> taxa = ['U1', 'U2', 'U3', 'U4', 'U5', 'U6', 'U7', 'U8']

    We also need a phylogenetic tree that relates the taxa to one another.

    >>> from io import StringIO
    >>> from skbio import TreeNode
    >>> tree = TreeNode.read(StringIO(
    ...                      '(((((U1:0.5,U2:0.5):0.5,U3:1.0):1.0):0.0,'
    ...                      '(U4:0.75,(U5:0.5,((U6:0.33,U7:0.62):0.5'
    ...                      ',U8:0.5):0.5):0.5):1.25):0.0)root;'))

    We can then compute the Faith PD of the sample.

    >>> from skbio.diversity.alpha import faith_pd
    >>> pd = faith_pd(u_counts, taxa, tree)
    >>> print(round(pd, 2))
    6.95

    """
    taxa = _check_taxa_alias(taxa, tree, otu_ids)

    counts_by_node, branch_lengths = _setup_pd(
        counts, taxa, tree, validate, rooted=True, single_sample=True
    )

    return _faith_pd(counts_by_node, branch_lengths)


def _phydiv(counts_by_node, branch_lengths, rooted, weight):
    """Calculate generalized phylogenetic diversity (PD) metrics.

    Parameters
    ----------
    counts_by_node : ndarray of shape (n_samples, n_nodes)
        Total counts/abundances of taxa descending from individual nodes of the tree.
    branch_lengths : ndarray of shape (n_nodes,)
        Branch lengths of corresponding nodes of the tree.
    rooted : bool
        Whether the metric is calculated considering the root of the tree.
    weight : bool or float
        Whether and to what degree branch lengths should be weighted by the relative
        abundance of taxa descending from the branch.

    Returns
    -------
    float
        Phylogenetic diversity (PD).

    """
    # select branches connecting taxa
    included = counts_by_node > 0

    # get total counts
    counts_sum = counts_by_node.max()
    if counts_sum == 0.0:
        return 0.0

    # in unrooted mode, remove branches to root
    if rooted is False:
        included &= counts_by_node < counts_sum

    # in unweighted mode, simply sum branch lengths
    if not weight:
        return (branch_lengths * included).sum()

    # get relative abundances
    fracs_by_node = counts_by_node / counts_sum

    # calculate balances in unrooted mode
    if rooted is False:
        fracs_by_node = 2 * np.minimum(fracs_by_node, 1 - fracs_by_node)

    # raise relative abundances to the power of theta
    if isinstance(weight, float) and weight < 1.0:
        fracs_by_node **= weight

    return (branch_lengths * fracs_by_node).sum()


def phydiv(
    counts, taxa=None, tree=None, rooted=None, weight=False, validate=True, otu_ids=None
):
    r"""Calculate generalized phylogenetic diversity (PD) metrics.

    Parameters
    ----------
    counts : 1-D array_like, int
        Vectors of counts/abundances of taxa for one sample.
    taxa : list, np.array
        Vector of taxon IDs corresponding to tip names in ``tree``. Must be the
        same length as ``counts``. Required.
    tree : skbio.TreeNode
        Tree relating taxa. The set of tip names in the tree can be a superset
        of ``taxa``, but not a subset. Required.
    rooted : bool, optional
        Whether the metric is calculated considering the root of the tree. By
        default, this will be determined based on whether the input tree is
        rooted. However, one can override it by explicitly specifying ``True``
        (rooted) or ``False`` (unrooted).
    weight : bool or float, optional
        Whether branch lengths should be weighted by the relative abundance of
        taxa descending from the branch (default: ``False``). A float within
        [0, 1] indicates the degree of partial-weighting (0: unweighted, 1:
        fully-weighted).
    validate: bool, optional
        Whether validate the input data. See ``faith_pd`` for details.
    otu_ids : list, np.array
        Alias of ``taxa`` for backward compatibility. Deprecated and to be
        removed in a future release.

    Returns
    -------
    float
        Phylogenetic diversity (PD).

    Raises
    ------
    ValueError, MissingNodeError, DuplicateNodeError
        If validation fails. Exact error will depend on what was invalid.

    See Also
    --------
    faith_pd
    skbio.diversity
    skbio.diversity.alpha_diversity

    Notes
    -----
    Phylogenetic diversity (PD) metrics measure the diversity of a community
    with consideration of the phylogenetic relationships among taxa. In
    general, PD is the sum of branch lengths spanning across taxa, optionally
    weighted by their abundance.

    The most widely-adopted PD metric, Faith's PD [1]_, is defined as:

    .. math::

       PD = \sum_{b \in T \sqcup R} l(b)

    where :math:`T` is a minimum set of branches (:math:`b`) in a rooted tree
    that connect all taxa in a community. :math:`R` is a set of branches from
    the lowest common ancestor (LCA) of the taxa to the root of the tree.
    :math:`PD` is the sum of lengths (:math:`l`) of branches in both sets.

    It is equivalent to ``pd(..., rooted=True, weight=False)``.

    A variant of PD, which does not include the root in the calculation, was
    referred to by some authors as unrooted phylogenetic diversity (uPD) [2]_,
    as in contrast to rooted phylogenetic diversity (rPD, i.e., Faith's PD).
    uPD is defined as:

    .. math::

       PD = \sum_{b \in T} l(b)

    It is equivalent to ``pd(..., rooted=False, weight=False)``.

    See ``faith_pd`` for a discussion of the root.

    PD (rooted or unrooted) considers only the presence of taxa. Therefore, it
    can be considered as the phylogenetic generalization of species richness.
    However, there are advantages of incorporating abundance information in the
    measurement [3]_. A generalized framework of abundance-weighted PD is
    provided in [4]_.

    Abundance-weighted rooted PD (equivalent to :math:`RBWPD_{1}` described in
    [4]_) is analogous to the :math:`PD_{aw}` metric originally described in
    [5]_ with a multiplier. It is defined as:

    .. math::

       PD = \sum_{b \in T \sqcup R} l(b) p(b)

    where :math:`p` is the sum of relative (proportional) abundances of taxa
    descending from branch (:math:`b`).

    It is equivalent to ``pd(..., rooted=True, weight=True)``.

    Abundance-weighted unrooted PD (equivalent to :math:`BWPD_{1}` described in
    [4]_) is analogous to the :math:`\delta nPD` metric originally described in
    [6]_ with a multiplier. It is defined as:

    .. math::

       PD = 2 \sum_{b \in T} l(b) \min(p(b),1-p(b))

    In which the term :math:`2\min(p(b),1-p(b))` is the lesser of the relative
    abundance of descending taxa on either side of a branch, multiplied by two.
    It is referred to as the "balance" of taxon abundance in [4]_.

    It is equivalent to ``pd(..., rooted=False, weight=True)``.

    The contribution of taxon abundance to the metric can be adjusted using the
    ``weighted`` parameter when it is a float within [0, 1]. This factor was
    referred to as :math:`\theta` in [4]_. The metric, :math:`BWPD_{\theta}`,
    referred to as the balance-weighted phylogenetic diversity in [4]_, is
    defined as:

    .. math::

       PD = \sum_{b \in T} l(b) (2\min(p(b),1-p(b)))^\theta

    It is equivalent to ``pd(..., rooted=False, weight=theta)``.

    This metric falls back to unweighted PD when :math:`\theta=0` or fully-
    weighted PD when :math:`\theta=1`. The original publication tested
    :math:`\theta=0.25` or :math:`0.5` [4]_.

    The parameter :math:`\theta` is analogous to the parameter :math:`\alpha`
    in the generalized UniFrac metric [7]_.

    Likewise, the rooted version of balance-weighted phylogenetic diversity,
    :math:`RBWPD_{\theta}` [4]_ (although "balance" is not involved), is
    defined as:

    .. math::

        PD = \sum_{b \in T \sqcup R} l(b) p(b)^\theta

    It is equivalent to ``pd(..., rooted=True, weight=theta)``.

    It is important to report which metric is used. For practical perspective,
    we recommend the following denotions:

    - :math:`rPD`: rooted, unweighted PD (Faith's PD [1]_).
    - :math:`uPD`: unrooted, unweighted PD (uPD [2]_).
    - :math:`rPD_{w}`: rooted, weighted PD (analogous to
      :math:`PD_{aw}` [3]_).
    - :math:`uPD_{w}`: unrooted, weighted PD (analogous to
      :math:`\delta nPD` [4]_).
    - :math:`rPD_{w\theta}`: rooted, weighted PD with parameter
      :math:`\theta` (:math:`RBWPD_{\theta}` [5]_).
    - :math:`uPD_{w\theta}`: unrooted, weighted PD with parameter
      :math:`\theta` (:math:`BWPD_{\theta}` [5]_).

    References
    ----------
    .. [1] Faith, D. P. Conservation evaluation and phylogenetic diversity.
       Biol. Conserv. (1992).

    .. [2] Pardi, F., & Goldman, N. (2007). Resource-aware taxon selection for
       maximizing phylogenetic diversity. Systematic biology, 56(3), 431-444.

    .. [3] Chao, A., Chiu, C. H., & Jost, L. (2016). Phylogenetic diversity
       measures and their decomposition: a framework based on Hill numbers.
       Biodiversity Conservation and Phylogenetic Systematics, 14, 141-72.

    .. [4] McCoy, C. O., & Matsen IV, F. A. (2013). Abundance-weighted
       phylogenetic diversity measures distinguish microbial community states
       and are robust to sampling depth. PeerJ, 1, e157.

    .. [5] Vellend, M., Cornwell, W. K., Magnuson-Ford, K., & Mooers, A. Ø.
       (2011). Measuring phylogenetic biodiversity. Biological diversity:
       frontiers in measurement and assessment, 194-207.

    .. [6] Barker, G. M. (2002). Phylogenetic diversity: a quantitative
       framework for measurement of priority and achievement in biodiversity
       conservation. Biological Journal of the Linnean Society, 76(2), 165-194.

    .. [7] Chen, J., Bittinger, K., Charlson, E. S., Hoffmann, C., Lewis, J.,
       Wu, G. D., ... & Li, H. (2012). Associating microbiome composition with
       environmental covariates using generalized UniFrac distances.
       Bioinformatics, 28(16), 2106-2113.

    """
    taxa = _check_taxa_alias(taxa, tree, otu_ids)

    # whether tree is rooted should not affect whether metric can be calculated
    # ; it is common unrooted PD is calculated on a rooted tree
    counts_by_node, branch_lengths = _setup_pd(
        counts, taxa, tree, validate, rooted=False, single_sample=True
    )

    # if not specified, determine whether metric should be calculated in rooted
    # mode according to the tree
    if rooted is None:
        rooted = len(tree.root().children) == 2

    # validate weight parameter
    if (
        not isinstance(weight, (bool, int, float))
        or (w_ := float(weight)) < 0.0
        or w_ > 1.0
    ):
        raise ValueError("Weight parameter must be boolean or within [0, 1].")

    return _phydiv(counts_by_node, branch_lengths, rooted, weight)



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

import copy

import pandas as pd
import numpy as np
import numpy.testing as npt

from skbio.util._testing import assert_data_frame_almost_equal
from skbio.metadata import IntervalMetadata


class MetadataMixinTests:
    def test_constructor_invalid_type(self):
        for md in (0, "a", ("f", "o", "o"), np.array([]), pd.DataFrame()):
            with self.assertRaisesRegex(TypeError, "metadata must be a dict"):
                self._metadata_constructor_(metadata=md)

    def test_constructor_no_metadata(self):
        for md in None, {}:
            obj = self._metadata_constructor_(metadata=md)

            self.assertFalse(obj.has_metadata())
            self.assertEqual(obj.metadata, {})

    def test_constructor_with_metadata(self):
        obj = self._metadata_constructor_(metadata={"foo": "bar"})
        self.assertEqual(obj.metadata, {"foo": "bar"})

        obj = self._metadata_constructor_(metadata={"": "", 123: {"a": "b", "c": "d"}})
        self.assertEqual(obj.metadata, {"": "", 123: {"a": "b", "c": "d"}})

    def test_constructor_handles_missing_metadata_efficiently(self):
        self.assertIsNone(self._metadata_constructor_()._metadata)
        self.assertIsNone(self._metadata_constructor_(metadata=None)._metadata)

    def test_constructor_makes_shallow_copy_of_metadata(self):
        md = {"foo": "bar", 42: []}
        obj = self._metadata_constructor_(metadata=md)

        self.assertEqual(obj.metadata, md)
        self.assertIsNot(obj.metadata, md)

        md["foo"] = "baz"
        self.assertEqual(obj.metadata, {"foo": "bar", 42: []})

        md[42].append(True)
        self.assertEqual(obj.metadata, {"foo": "bar", 42: [True]})

    def test_eq(self):
        self.assertReallyEqual(
            self._metadata_constructor_(metadata={"foo": 42}),
            self._metadata_constructor_(metadata={"foo": 42}),
        )

        self.assertReallyEqual(
            self._metadata_constructor_(metadata={"foo": 42, 123: {}}),
            self._metadata_constructor_(metadata={"foo": 42, 123: {}}),
        )

    def test_eq_missing_metadata(self):
        self.assertReallyEqual(
            self._metadata_constructor_(), self._metadata_constructor_()
        )
        self.assertReallyEqual(
            self._metadata_constructor_(), self._metadata_constructor_(metadata={})
        )
        self.assertReallyEqual(
            self._metadata_constructor_(metadata={}),
            self._metadata_constructor_(metadata={}),
        )

    def test_eq_handles_missing_metadata_efficiently(self):
        obj1 = self._metadata_constructor_()
        obj2 = self._metadata_constructor_()
        self.assertReallyEqual(obj1, obj2)

        self.assertIsNone(obj1._metadata)
        self.assertIsNone(obj2._metadata)

    def test_ne(self):
        # Both have metadata.
        obj1 = self._metadata_constructor_(metadata={"id": "foo"})
        obj2 = self._metadata_constructor_(metadata={"id": "bar"})
        self.assertReallyNotEqual(obj1, obj2)

        # One has metadata.
        obj1 = self._metadata_constructor_(metadata={"id": "foo"})
        obj2 = self._metadata_constructor_()
        self.assertReallyNotEqual(obj1, obj2)

    def test_copy_metadata_none(self):
        obj = self._metadata_constructor_()
        obj_copy = copy.copy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNone(obj._metadata)
        self.assertIsNone(obj_copy._metadata)

    def test_copy_metadata_empty(self):
        obj = self._metadata_constructor_(metadata={})
        obj_copy = copy.copy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertEqual(obj._metadata, {})
        self.assertIsNone(obj_copy._metadata)

    def test_copy_with_metadata(self):
        obj = self._metadata_constructor_(metadata={"foo": [1]})
        obj_copy = copy.copy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNot(obj._metadata, obj_copy._metadata)
        self.assertIs(obj._metadata["foo"], obj_copy._metadata["foo"])

        obj_copy.metadata["foo"].append(2)
        obj_copy.metadata["foo2"] = 42

        self.assertEqual(obj_copy.metadata, {"foo": [1, 2], "foo2": 42})
        self.assertEqual(obj.metadata, {"foo": [1, 2]})

    def test_deepcopy_metadata_none(self):
        obj = self._metadata_constructor_()
        obj_copy = copy.deepcopy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNone(obj._metadata)
        self.assertIsNone(obj_copy._metadata)

    def test_deepcopy_metadata_empty(self):
        obj = self._metadata_constructor_(metadata={})
        obj_copy = copy.deepcopy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertEqual(obj._metadata, {})
        self.assertIsNone(obj_copy._metadata)

    def test_deepcopy_with_metadata(self):
        obj = self._metadata_constructor_(metadata={"foo": [1]})
        obj_copy = copy.deepcopy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNot(obj._metadata, obj_copy._metadata)
        self.assertIsNot(obj._metadata["foo"], obj_copy._metadata["foo"])

        obj_copy.metadata["foo"].append(2)
        obj_copy.metadata["foo2"] = 42

        self.assertEqual(obj_copy.metadata, {"foo": [1, 2], "foo2": 42})
        self.assertEqual(obj.metadata, {"foo": [1]})

    def test_deepcopy_memo_is_respected(self):
        # Basic test to ensure deepcopy's memo is passed through to recursive
        # deepcopy calls.
        obj = self._metadata_constructor_(metadata={"foo": "bar"})
        memo = {}
        copy.deepcopy(obj, memo)
        self.assertGreater(len(memo), 2)

    def test_metadata_getter(self):
        obj = self._metadata_constructor_(metadata={42: "foo", ("hello", "world"): 43})

        self.assertIsInstance(obj.metadata, dict)
        self.assertEqual(obj.metadata, {42: "foo", ("hello", "world"): 43})

        obj.metadata[42] = "bar"
        self.assertEqual(obj.metadata, {42: "bar", ("hello", "world"): 43})

    def test_metadata_getter_no_metadata(self):
        obj = self._metadata_constructor_()

        self.assertIsNone(obj._metadata)
        self.assertIsInstance(obj.metadata, dict)
        self.assertEqual(obj.metadata, {})
        self.assertIsNotNone(obj._metadata)

    def test_metadata_setter(self):
        obj = self._metadata_constructor_()
        self.assertFalse(obj.has_metadata())

        obj.metadata = {"hello": "world"}
        self.assertTrue(obj.has_metadata())
        self.assertEqual(obj.metadata, {"hello": "world"})

        obj.metadata = {}
        self.assertFalse(obj.has_metadata())
        self.assertEqual(obj.metadata, {})

    def test_metadata_setter_makes_shallow_copy(self):
        obj = self._metadata_constructor_()

        md = {"foo": "bar", 42: []}
        obj.metadata = md

        self.assertEqual(obj.metadata, md)
        self.assertIsNot(obj.metadata, md)

        md["foo"] = "baz"
        self.assertEqual(obj.metadata, {"foo": "bar", 42: []})

        md[42].append(True)
        self.assertEqual(obj.metadata, {"foo": "bar", 42: [True]})

    def test_metadata_setter_invalid_type(self):
        obj = self._metadata_constructor_(metadata={123: 456})

        for md in (None, 0, "a", ("f", "o", "o"), np.array([]), pd.DataFrame()):
            with self.assertRaisesRegex(TypeError, "metadata must be a dict"):
                obj.metadata = md
            self.assertEqual(obj.metadata, {123: 456})

    def test_metadata_deleter(self):
        obj = self._metadata_constructor_(metadata={"foo": "bar"})

        self.assertEqual(obj.metadata, {"foo": "bar"})

        del obj.metadata
        self.assertIsNone(obj._metadata)
        self.assertFalse(obj.has_metadata())

        # Delete again.
        del obj.metadata
        self.assertIsNone(obj._metadata)
        self.assertFalse(obj.has_metadata())

        obj = self._metadata_constructor_()

        self.assertIsNone(obj._metadata)
        self.assertFalse(obj.has_metadata())
        del obj.metadata
        self.assertIsNone(obj._metadata)
        self.assertFalse(obj.has_metadata())

    def test_has_metadata(self):
        obj = self._metadata_constructor_()

        self.assertFalse(obj.has_metadata())
        # Handles metadata efficiently.
        self.assertIsNone(obj._metadata)

        self.assertFalse(self._metadata_constructor_(metadata={}).has_metadata())

        self.assertTrue(self._metadata_constructor_(metadata={"": ""}).has_metadata())
        self.assertTrue(
            self._metadata_constructor_(metadata={"foo": 42}).has_metadata()
        )


class PositionalMetadataMixinTests:
    def test_constructor_invalid_positional_metadata_type(self):
        with self.assertRaisesRegex(
            TypeError,
            "Invalid positional metadata. Must be "
            "consumable by `pd.DataFrame` constructor."
            " Original pandas error message: ",
        ):
            self._positional_metadata_constructor_(0, positional_metadata=2)

    def test_constructor_positional_metadata_len_mismatch(self):
        # Zero elements.
        with self.assertRaisesRegex(ValueError, r"\(0\).*\(4\)"):
            self._positional_metadata_constructor_(4, positional_metadata=[])

        # Not enough elements.
        with self.assertRaisesRegex(ValueError, r"\(3\).*\(4\)"):
            self._positional_metadata_constructor_(4, positional_metadata=[2, 3, 4])

        # Too many elements.
        with self.assertRaisesRegex(ValueError, r"\(5\).*\(4\)"):
            self._positional_metadata_constructor_(
                4, positional_metadata=[2, 3, 4, 5, 6]
            )

        # Series not enough rows.
        with self.assertRaisesRegex(ValueError, r"\(3\).*\(4\)"):
            self._positional_metadata_constructor_(
                4, positional_metadata=pd.Series(range(3))
            )

        # Series too many rows.
        with self.assertRaisesRegex(ValueError, r"\(5\).*\(4\)"):
            self._positional_metadata_constructor_(
                4, positional_metadata=pd.Series(range(5))
            )

        # DataFrame not enough rows.
        with self.assertRaisesRegex(ValueError, r"\(3\).*\(4\)"):
            self._positional_metadata_constructor_(
                4, positional_metadata=pd.DataFrame({"quality": range(3)})
            )

        # DataFrame too many rows.
        with self.assertRaisesRegex(ValueError, r"\(5\).*\(4\)"):
            self._positional_metadata_constructor_(
                4, positional_metadata=pd.DataFrame({"quality": range(5)})
            )

        # Empty DataFrame wrong size.
        with self.assertRaisesRegex(ValueError, r"\(2\).*\(3\)"):
            self._positional_metadata_constructor_(
                3, positional_metadata=pd.DataFrame(index=range(2))
            )

    def test_constructor_no_positional_metadata(self):
        # Length zero with missing/empty positional metadata.
        for empty in None, {}, pd.DataFrame():
            obj = self._positional_metadata_constructor_(0, positional_metadata=empty)

            self.assertFalse(obj.has_positional_metadata())
            self.assertIsInstance(obj.positional_metadata.index, pd.RangeIndex)
            assert_data_frame_almost_equal(
                obj.positional_metadata, pd.DataFrame(index=range(0))
            )

        # Nonzero length with missing positional metadata.
        obj = self._positional_metadata_constructor_(3, positional_metadata=None)

        self.assertFalse(obj.has_positional_metadata())
        self.assertIsInstance(obj.positional_metadata.index, pd.RangeIndex)
        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame(index=range(3))
        )

    def test_constructor_with_positional_metadata_len_zero(self):
        for data in [], (), np.array([]):
            obj = self._positional_metadata_constructor_(
                0, positional_metadata={"foo": data}
            )

            self.assertTrue(obj.has_positional_metadata())
            assert_data_frame_almost_equal(
                obj.positional_metadata, pd.DataFrame({"foo": data}, index=range(0))
            )

    def test_constructor_with_positional_metadata_len_one(self):
        for data in [2], (2,), np.array([2]):
            obj = self._positional_metadata_constructor_(
                1, positional_metadata={"foo": data}
            )

            self.assertTrue(obj.has_positional_metadata())
            assert_data_frame_almost_equal(
                obj.positional_metadata, pd.DataFrame({"foo": data}, index=range(1))
            )

    def test_constructor_with_positional_metadata_len_greater_than_one(self):
        for data in (
            [0, 42, 42, 1, 0, 8, 100, 0, 0],
            (0, 42, 42, 1, 0, 8, 100, 0, 0),
            np.array([0, 42, 42, 1, 0, 8, 100, 0, 0]),
        ):
            obj = self._positional_metadata_constructor_(
                9, positional_metadata={"foo": data}
            )

            self.assertTrue(obj.has_positional_metadata())
            assert_data_frame_almost_equal(
                obj.positional_metadata, pd.DataFrame({"foo": data}, index=range(9))
            )

    def test_constructor_with_positional_metadata_multiple_columns(self):
        obj = self._positional_metadata_constructor_(
            5, positional_metadata={"foo": np.arange(5), "bar": np.arange(5)[::-1]}
        )

        self.assertTrue(obj.has_positional_metadata())
        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame(
                {"foo": np.arange(5), "bar": np.arange(5)[::-1]}, index=range(5)
            ),
        )

    def test_constructor_with_positional_metadata_custom_index(self):
        df = pd.DataFrame(
            {"foo": np.arange(5), "bar": np.arange(5)[::-1]},
            index=["a", "b", "c", "d", "e"],
        )
        obj = self._positional_metadata_constructor_(5, positional_metadata=df)

        self.assertTrue(obj.has_positional_metadata())
        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame(
                {"foo": np.arange(5), "bar": np.arange(5)[::-1]}, index=range(5)
            ),
        )

    def test_constructor_with_positional_metadata_int64_index(self):
        # Test that memory-inefficient index is converted to memory-efficient
        # index.
        df = pd.DataFrame(
            {"foo": np.arange(5), "bar": np.arange(5)[::-1]}, index=np.arange(5)
        )
        self.assertEqual(df.index.dtype, np.int_)

        obj = self._positional_metadata_constructor_(5, positional_metadata=df)

        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame(
                {"foo": np.arange(5), "bar": np.arange(5)[::-1]}, index=range(5)
            ),
        )
        self.assertIsInstance(obj.positional_metadata.index, pd.RangeIndex)

    def test_constructor_handles_missing_positional_metadata_efficiently(self):
        obj = self._positional_metadata_constructor_(4)
        self.assertIsNone(obj._positional_metadata)

        obj = self._positional_metadata_constructor_(4, positional_metadata=None)
        self.assertIsNone(obj._positional_metadata)

    def test_constructor_makes_shallow_copy_of_positional_metadata(self):
        df = pd.DataFrame(
            {"foo": [22, 22, 0], "bar": [[], [], []]}, index=["a", "b", "c"]
        )
        obj = self._positional_metadata_constructor_(3, positional_metadata=df)

        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame({"foo": [22, 22, 0], "bar": [[], [], []]}, index=range(3)),
        )
        self.assertIsNot(obj.positional_metadata, df)

        # Original df is not mutated.
        orig_df = pd.DataFrame(
            {"foo": [22, 22, 0], "bar": [[], [], []]}, index=["a", "b", "c"]
        )
        assert_data_frame_almost_equal(df, orig_df)

        # Change values of column (using same dtype).
        df["foo"] = [42, 42, 42]
        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame({"foo": [22, 22, 0], "bar": [[], [], []]}, index=range(3)),
        )

        # Change single value of underlying data.
        df.values[0][0] = 10
        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame({"foo": [22, 22, 0], "bar": [[], [], []]}, index=range(3)),
        )

        # Mutate list (not a deep copy).
        df["bar"].iloc[0].append(42)
        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame({"foo": [22, 22, 0], "bar": [[42], [], []]}, index=range(3)),
        )

    def test_eq_basic(self):
        obj1 = self._positional_metadata_constructor_(
            3, positional_metadata={"foo": [1, 2, 3]}
        )
        obj2 = self._positional_metadata_constructor_(
            3, positional_metadata={"foo": [1, 2, 3]}
        )
        self.assertReallyEqual(obj1, obj2)

    def test_eq_from_different_source(self):
        obj1 = self._positional_metadata_constructor_(
            3, positional_metadata={"foo": np.array([1, 2, 3], dtype=np.int_)}
        )
        obj2 = self._positional_metadata_constructor_(
            3,
            positional_metadata=pd.DataFrame(
                {"foo": [1, 2, 3]}, index=["foo", "bar", "baz"], dtype=np.int_
            ),
        )
        self.assertReallyEqual(obj1, obj2)

    def test_eq_missing_positional_metadata(self):
        for empty in None, {}, pd.DataFrame(), pd.DataFrame(index=[]):
            obj = self._positional_metadata_constructor_(0, positional_metadata=empty)

            self.assertReallyEqual(obj, self._positional_metadata_constructor_(0))
            self.assertReallyEqual(
                obj,
                self._positional_metadata_constructor_(0, positional_metadata=empty),
            )

        for empty in None, pd.DataFrame(index=["a", "b"]):
            obj = self._positional_metadata_constructor_(2, positional_metadata=empty)

            self.assertReallyEqual(obj, self._positional_metadata_constructor_(2))
            self.assertReallyEqual(
                obj,
                self._positional_metadata_constructor_(2, positional_metadata=empty),
            )

    def test_eq_handles_missing_positional_metadata_efficiently(self):
        obj1 = self._positional_metadata_constructor_(1)
        obj2 = self._positional_metadata_constructor_(1)
        self.assertReallyEqual(obj1, obj2)

        self.assertIsNone(obj1._positional_metadata)
        self.assertIsNone(obj2._positional_metadata)

    def test_ne_len_zero(self):
        # Both have positional metadata.
        obj1 = self._positional_metadata_constructor_(
            0, positional_metadata={"foo": []}
        )
        obj2 = self._positional_metadata_constructor_(
            0, positional_metadata={"foo": [], "bar": []}
        )
        self.assertReallyNotEqual(obj1, obj2)

        # One has positional metadata.
        obj1 = self._positional_metadata_constructor_(
            0, positional_metadata={"foo": []}
        )
        obj2 = self._positional_metadata_constructor_(0)
        self.assertReallyNotEqual(obj1, obj2)

    def test_ne_len_greater_than_zero(self):
        # Both have positional metadata.
        obj1 = self._positional_metadata_constructor_(
            3, positional_metadata={"foo": [1, 2, 3]}
        )
        obj2 = self._positional_metadata_constructor_(
            3, positional_metadata={"foo": [1, 2, 2]}
        )
        self.assertReallyNotEqual(obj1, obj2)

        # One has positional metadata.
        obj1 = self._positional_metadata_constructor_(
            3, positional_metadata={"foo": [1, 2, 3]}
        )
        obj2 = self._positional_metadata_constructor_(3)
        self.assertReallyNotEqual(obj1, obj2)

    def test_ne_len_mismatch(self):
        obj1 = self._positional_metadata_constructor_(
            3, positional_metadata=pd.DataFrame(index=range(3))
        )
        obj2 = self._positional_metadata_constructor_(
            2, positional_metadata=pd.DataFrame(index=range(2))
        )
        self.assertReallyNotEqual(obj1, obj2)

    def test_copy_positional_metadata_none(self):
        obj = self._positional_metadata_constructor_(3)
        obj_copy = copy.copy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNone(obj._positional_metadata)
        self.assertIsNone(obj_copy._positional_metadata)

    def test_copy_positional_metadata_empty(self):
        obj = self._positional_metadata_constructor_(
            3, positional_metadata=pd.DataFrame(index=range(3))
        )
        obj_copy = copy.copy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        assert_data_frame_almost_equal(
            obj._positional_metadata, pd.DataFrame(index=range(3))
        )
        self.assertIsNone(obj_copy._positional_metadata)

    def test_copy_with_positional_metadata(self):
        obj = self._positional_metadata_constructor_(
            4, positional_metadata={"bar": [[], [], [], []], "baz": [42, 42, 42, 42]}
        )
        obj_copy = copy.copy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNot(obj._positional_metadata, obj_copy._positional_metadata)
        self.assertIsNot(
            obj._positional_metadata.values, obj_copy._positional_metadata.values
        )
        self.assertIs(
            obj._positional_metadata.loc[0, "bar"],
            obj_copy._positional_metadata.loc[0, "bar"],
        )

        obj_copy.positional_metadata.loc[0, "bar"].append(1)
        obj_copy.positional_metadata.loc[0, "baz"] = 43

        assert_data_frame_almost_equal(
            obj_copy.positional_metadata,
            pd.DataFrame({"bar": [[1], [], [], []], "baz": [43, 42, 42, 42]}),
        )
        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame({"bar": [[1], [], [], []], "baz": [42, 42, 42, 42]}),
        )

    def test_copy_preserves_range_index(self):
        for pm in None, {"foo": ["a", "b", "c"]}:
            obj = self._positional_metadata_constructor_(3, positional_metadata=pm)
            obj_copy = copy.copy(obj)

            self.assertIsInstance(obj.positional_metadata.index, pd.RangeIndex)
            self.assertIsInstance(obj_copy.positional_metadata.index, pd.RangeIndex)

    def test_deepcopy_positional_metadata_none(self):
        obj = self._positional_metadata_constructor_(3)
        obj_copy = copy.deepcopy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNone(obj._positional_metadata)
        self.assertIsNone(obj_copy._positional_metadata)

    def test_deepcopy_positional_metadata_empty(self):
        obj = self._positional_metadata_constructor_(
            3, positional_metadata=pd.DataFrame(index=range(3))
        )
        obj_copy = copy.deepcopy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        assert_data_frame_almost_equal(
            obj._positional_metadata, pd.DataFrame(index=range(3))
        )
        self.assertIsNone(obj_copy._positional_metadata)

    def test_deepcopy_with_positional_metadata(self):
        obj = self._positional_metadata_constructor_(
            4, positional_metadata={"bar": [[], [], [], []], "baz": [42, 42, 42, 42]}
        )
        obj_copy = copy.deepcopy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNot(obj._positional_metadata, obj_copy._positional_metadata)
        self.assertIsNot(
            obj._positional_metadata.values, obj_copy._positional_metadata.values
        )
        self.assertIsNot(
            obj._positional_metadata.loc[0, "bar"],
            obj_copy._positional_metadata.loc[0, "bar"],
        )

        obj_copy.positional_metadata.loc[0, "bar"].append(1)
        obj_copy.positional_metadata.loc[0, "baz"] = 43

        assert_data_frame_almost_equal(
            obj_copy.positional_metadata,
            pd.DataFrame({"bar": [[1], [], [], []], "baz": [43, 42, 42, 42]}),
        )
        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame({"bar": [[], [], [], []], "baz": [42, 42, 42, 42]}),
        )

    def test_deepcopy_preserves_range_index(self):
        for pm in None, {"foo": ["a", "b", "c"]}:
            obj = self._positional_metadata_constructor_(3, positional_metadata=pm)
            obj_copy = copy.deepcopy(obj)

            self.assertIsInstance(obj.positional_metadata.index, pd.RangeIndex)
            self.assertIsInstance(obj_copy.positional_metadata.index, pd.RangeIndex)

    def test_deepcopy_memo_is_respected(self):
        # Basic test to ensure deepcopy's memo is passed through to recursive
        # deepcopy calls.
        obj = self._positional_metadata_constructor_(
            3, positional_metadata={"foo": [1, 2, 3]}
        )
        memo = {}
        copy.deepcopy(obj, memo)
        self.assertGreater(len(memo), 2)

    def test_positional_metadata_getter(self):
        obj = self._positional_metadata_constructor_(
            3, positional_metadata={"foo": [22, 22, 0]}
        )

        self.assertIsInstance(obj.positional_metadata, pd.DataFrame)
        self.assertIsInstance(obj.positional_metadata.index, pd.RangeIndex)
        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame({"foo": [22, 22, 0]})
        )

        # Update existing column.
        obj.positional_metadata["foo"] = [42, 42, 43]
        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame({"foo": [42, 42, 43]})
        )

        # Add new column.
        obj.positional_metadata["foo2"] = [True, False, True]
        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame({"foo": [42, 42, 43], "foo2": [True, False, True]}),
        )

    def test_positional_metadata_getter_no_positional_metadata(self):
        obj = self._positional_metadata_constructor_(4)

        self.assertIsNone(obj._positional_metadata)
        self.assertIsInstance(obj.positional_metadata, pd.DataFrame)
        self.assertIsInstance(obj.positional_metadata.index, pd.RangeIndex)
        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame(index=range(4))
        )
        self.assertIsNotNone(obj._positional_metadata)

    def test_positional_metadata_getter_set_column_series(self):
        length = 8
        obj = self._positional_metadata_constructor_(
            length, positional_metadata={"foo": range(length)}
        )

        obj.positional_metadata["bar"] = pd.Series(range(length - 3))
        # pandas.Series will be padded with NaN if too short.
        npt.assert_equal(
            obj.positional_metadata["bar"],
            np.array(list(range(length - 3)) + [np.nan] * 3),
        )

        obj.positional_metadata["baz"] = pd.Series(range(length + 3))
        # pandas.Series will be truncated if too long.
        npt.assert_equal(obj.positional_metadata["baz"], np.array(range(length)))

    def test_positional_metadata_getter_set_column_array(self):
        length = 8
        obj = self._positional_metadata_constructor_(
            length, positional_metadata={"foo": range(length)}
        )

        # array-like objects will fail if wrong size.
        for array_like in (
            np.array(range(length - 1)),
            range(length - 1),
            np.array(range(length + 1)),
            range(length + 1),
        ):
            with self.assertRaisesRegex(
                ValueError,
                r"Length of values \("
                + str(len(array_like))
                + r"\) does not match length"
                r" of index \(8\)",
            ):
                obj.positional_metadata["bar"] = array_like

    def test_positional_metadata_setter_pandas_consumable(self):
        obj = self._positional_metadata_constructor_(3)

        self.assertFalse(obj.has_positional_metadata())

        obj.positional_metadata = {"foo": [3, 2, 1]}
        self.assertTrue(obj.has_positional_metadata())
        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame({"foo": [3, 2, 1]})
        )

        obj.positional_metadata = pd.DataFrame(index=np.arange(3))
        self.assertFalse(obj.has_positional_metadata())
        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame(index=range(3))
        )

    def test_positional_metadata_setter_data_frame(self):
        obj = self._positional_metadata_constructor_(3)

        self.assertFalse(obj.has_positional_metadata())

        obj.positional_metadata = pd.DataFrame(
            {"foo": [3, 2, 1]}, index=["a", "b", "c"]
        )
        self.assertTrue(obj.has_positional_metadata())
        self.assertIsInstance(obj.positional_metadata.index, pd.RangeIndex)
        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame({"foo": [3, 2, 1]})
        )

        obj.positional_metadata = pd.DataFrame(index=np.arange(3))
        self.assertFalse(obj.has_positional_metadata())
        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame(index=range(3))
        )

    def test_positional_metadata_setter_none(self):
        obj = self._positional_metadata_constructor_(0, positional_metadata={"foo": []})

        self.assertTrue(obj.has_positional_metadata())
        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame({"foo": []})
        )

        # `None` behavior differs from constructor.
        obj.positional_metadata = None

        self.assertFalse(obj.has_positional_metadata())
        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame(index=range(0))
        )

    def test_positional_metadata_setter_int64_index(self):
        # Test that memory-inefficient index is converted to memory-efficient
        # index.
        obj = self._positional_metadata_constructor_(5)

        df = pd.DataFrame(
            {"foo": np.arange(5), "bar": np.arange(5)[::-1]}, index=np.arange(5)
        )
        self.assertEqual(df.index.dtype, np.int_)

        obj.positional_metadata = df

        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame(
                {"foo": np.arange(5), "bar": np.arange(5)[::-1]}, index=range(5)
            ),
        )
        self.assertIsInstance(obj.positional_metadata.index, pd.RangeIndex)

    def test_positional_metadata_setter_makes_shallow_copy(self):
        obj = self._positional_metadata_constructor_(3)

        df = pd.DataFrame(
            {"foo": [22, 22, 0], "bar": [[], [], []]}, index=["a", "b", "c"]
        )
        obj.positional_metadata = df

        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame({"foo": [22, 22, 0], "bar": [[], [], []]}, index=range(3)),
        )
        self.assertIsNot(obj.positional_metadata, df)

        # Original df is not mutated.
        orig_df = pd.DataFrame(
            {"foo": [22, 22, 0], "bar": [[], [], []]}, index=["a", "b", "c"]
        )
        assert_data_frame_almost_equal(df, orig_df)

        # Change values of column (using same dtype).
        df["foo"] = [42, 42, 42]
        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame({"foo": [22, 22, 0], "bar": [[], [], []]}, index=range(3)),
        )

        # Change single value of underlying data.
        df.values[0][0] = 10
        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame({"foo": [22, 22, 0], "bar": [[], [], []]}, index=range(3)),
        )

        # Mutate list (not a deep copy).
        df["bar"].iloc[0].append(42)
        assert_data_frame_almost_equal(
            obj.positional_metadata,
            pd.DataFrame({"foo": [22, 22, 0], "bar": [[42], [], []]}, index=range(3)),
        )

    def test_positional_metadata_setter_invalid_type(self):
        obj = self._positional_metadata_constructor_(
            3, positional_metadata={"foo": [1, 2, 42]}
        )

        with self.assertRaisesRegex(
            TypeError,
            "Invalid positional metadata. Must be "
            "consumable by `pd.DataFrame` constructor."
            " Original pandas error message: ",
        ):
            obj.positional_metadata = 2

        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame({"foo": [1, 2, 42]})
        )

    def test_positional_metadata_setter_len_mismatch(self):
        obj = self._positional_metadata_constructor_(
            3, positional_metadata={"foo": [1, 2, 42]}
        )

        # `None` behavior differs from constructor.
        with self.assertRaisesRegex(ValueError, r"\(0\).*\(3\)"):
            obj.positional_metadata = None

        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame({"foo": [1, 2, 42]})
        )

        with self.assertRaisesRegex(ValueError, r"\(4\).*\(3\)"):
            obj.positional_metadata = [1, 2, 3, 4]

        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame({"foo": [1, 2, 42]})
        )

    def test_positional_metadata_deleter(self):
        obj = self._positional_metadata_constructor_(
            3, positional_metadata={"foo": [1, 2, 3]}
        )

        self.assertIsInstance(obj.positional_metadata.index, pd.RangeIndex)
        assert_data_frame_almost_equal(
            obj.positional_metadata, pd.DataFrame({"foo": [1, 2, 3]})
        )

        del obj.positional_metadata
        self.assertIsNone(obj._positional_metadata)
        self.assertFalse(obj.has_positional_metadata())

        # Delete again.
        del obj.positional_metadata
        self.assertIsNone(obj._positional_metadata)
        self.assertFalse(obj.has_positional_metadata())

        obj = self._positional_metadata_constructor_(3)

        self.assertIsNone(obj._positional_metadata)
        self.assertFalse(obj.has_positional_metadata())
        del obj.positional_metadata
        self.assertIsNone(obj._positional_metadata)
        self.assertFalse(obj.has_positional_metadata())

    def test_has_positional_metadata(self):
        obj = self._positional_metadata_constructor_(4)
        self.assertFalse(obj.has_positional_metadata())
        self.assertIsNone(obj._positional_metadata)

        obj = self._positional_metadata_constructor_(0, positional_metadata={})
        self.assertFalse(obj.has_positional_metadata())

        obj = self._positional_metadata_constructor_(
            4, positional_metadata=pd.DataFrame(index=np.arange(4))
        )
        self.assertFalse(obj.has_positional_metadata())

        obj = self._positional_metadata_constructor_(
            4, positional_metadata=pd.DataFrame(index=["a", "b", "c", "d"])
        )
        self.assertFalse(obj.has_positional_metadata())

        obj = self._positional_metadata_constructor_(0, positional_metadata={"foo": []})
        self.assertTrue(obj.has_positional_metadata())

        obj = self._positional_metadata_constructor_(
            4, positional_metadata={"foo": [1, 2, 3, 4]}
        )
        self.assertTrue(obj.has_positional_metadata())

        obj = self._positional_metadata_constructor_(
            2, positional_metadata={"foo": [1, 2], "bar": ["abc", "def"]}
        )
        self.assertTrue(obj.has_positional_metadata())


class IntervalMetadataMixinTests:
    def _set_up(self):
        self.upper_bound = 9
        self.im = IntervalMetadata(self.upper_bound)
        self.intvls = [
            {"bounds": [(0, 1), (2, 9)], "metadata": {"gene": "sagA"}},
            {"bounds": [(0, 1)], "metadata": {"gene": ["a"], "product": "foo"}},
        ]

    def test_constructor_invalid(self):
        with self.assertRaisesRegex(
            TypeError, "You must provide `IntervalMetadata` " "object."
        ):
            self._interval_metadata_constructor_(0, "")

    def test_constructor_empty_interval_metadata_upper_bound_is_none(self):
        im = IntervalMetadata(None)
        for i in [0, 1, 3, 100]:
            x = self._interval_metadata_constructor_(i, im)
            # the upper bound is reset to seq/axis length
            self.assertEqual(x.interval_metadata.upper_bound, i)
            self.assertEqual(x.interval_metadata._intervals, im._intervals)
            # original interval metadata upper bound is not changed
            self.assertIsNone(im.upper_bound)

    def test_constructor_interval_metadata_upper_bound_is_none(self):
        im = IntervalMetadata(None)
        # populate im
        im.add(**self.intvls[0])
        im.add(**self.intvls[1])
        for i in [1000, 100]:
            x = self._interval_metadata_constructor_(i, im)
            # the upper bound is reset to seq/axis length
            self.assertEqual(x.interval_metadata.upper_bound, i)
            self.assertEqual(x.interval_metadata._intervals, im._intervals)
            # original interval metadata upper bound is not changed
            self.assertIsNone(im.upper_bound)

    def test_constructor_interval_bounds_larger_than_len(self):
        im = IntervalMetadata(None)
        # populate im
        im.add(**self.intvls[0])
        im.add(**self.intvls[1])
        for i in [0, 1, 3]:
            # error to reset upper bound to a smaller value than seq/axis len
            with self.assertRaisesRegex(
                ValueError, r"larger than upper bound \(%r\)" % i
            ):
                self._interval_metadata_constructor_(i, im)
            # original interval metadata upper bound is not changed
            self.assertIsNone(im.upper_bound)

    def test_constructor_interval_metadata_len_mismatch(self):
        for i in [0, 1, 3, 100]:
            with self.assertRaisesRegex(
                ValueError, r"\(%d\).*\(%d\)" % (self.upper_bound, i)
            ):
                self._interval_metadata_constructor_(i, self.im)

    def test_constructor_interval_metadata_len(self):
        for n in 1, 2, 3:
            im = IntervalMetadata(n)
            im.add([(0, 1)], metadata={"a": "b"})
            obj = self._interval_metadata_constructor_(n, im)
            self.assertTrue(obj.has_interval_metadata())
            self.assertIsInstance(obj.interval_metadata, IntervalMetadata)

    def test_constructor_interval_metadata_len_0(self):
        im = IntervalMetadata(0)
        obj = self._interval_metadata_constructor_(0, im)
        self.assertFalse(obj.has_interval_metadata())

    def test_constructor_no_interval_metadata(self):
        for i, im in [(0, None), (self.upper_bound, self.im)]:
            obj = self._interval_metadata_constructor_(i, im)
            self.assertFalse(obj.has_interval_metadata())
            self.assertIsInstance(obj.interval_metadata, IntervalMetadata)

    def test_constructor_handles_missing_interval_metadata_efficiently(self):
        obj = self._interval_metadata_constructor_(self.upper_bound)
        self.assertIsNone(obj._interval_metadata)

        obj = self._interval_metadata_constructor_(
            self.upper_bound, interval_metadata=None
        )
        self.assertIsNone(obj._interval_metadata)

    def test_constructor_makes_shallow_copy_of_interval_metadata(self):
        intvl = self.im.add(**self.intvls[1])
        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)

        self.assertEqual(obj.interval_metadata, self.im)
        self.assertIsNot(obj.interval_metadata, self.im)

        # Changing mutable value of metadata of the old interval
        # also changes obj.
        intvl.metadata["gene"].append("b")
        self.assertEqual(obj.interval_metadata, self.im)

        # Changing old interval doesn't change obj
        intvl.bounds = [(3, 6)]
        self.assertNotEqual(obj.interval_metadata, self.im)

    def test_eq_basic(self):
        im1 = IntervalMetadata(self.upper_bound)
        im1.add(**self.intvls[0])
        obj1 = self._interval_metadata_constructor_(self.upper_bound, im1)

        im2 = IntervalMetadata(self.upper_bound)
        im2.add(**self.intvls[0])
        obj2 = self._interval_metadata_constructor_(self.upper_bound, im2)

        self.assertReallyEqual(obj1, obj2)

    def test_eq_populated_differently(self):
        im1 = IntervalMetadata(self.upper_bound)
        im1.add(**self.intvls[0])
        obj1 = self._interval_metadata_constructor_(self.upper_bound, im1)

        obj2 = self._interval_metadata_constructor_(self.upper_bound)
        obj2.interval_metadata.add(**self.intvls[0])

        self.assertReallyEqual(obj1, obj2)

    def test_eq_handles_missing_positional_metadata_efficiently(self):
        obj1 = self._interval_metadata_constructor_(self.upper_bound)
        obj2 = self._interval_metadata_constructor_(self.upper_bound)
        self.assertReallyEqual(obj1, obj2)

        self.assertIsNone(obj1._interval_metadata)
        self.assertIsNone(obj2._interval_metadata)

    def test_ne_diff_len(self):
        obj1 = self._interval_metadata_constructor_(0)
        obj2 = self._interval_metadata_constructor_(self.upper_bound)
        self.assertReallyNotEqual(obj1, obj2)

    def test_ne_only_one_is_empty(self):
        im1 = IntervalMetadata(self.upper_bound)
        im1.add(**self.intvls[0])
        obj1 = self._interval_metadata_constructor_(self.upper_bound, im1)

        obj2 = self._interval_metadata_constructor_(self.upper_bound)

        self.assertReallyNotEqual(obj1, obj2)

    def test_ne(self):
        im1 = IntervalMetadata(self.upper_bound)
        im1.add(**self.intvls[0])
        obj1 = self._interval_metadata_constructor_(self.upper_bound, im1)

        im2 = IntervalMetadata(self.upper_bound)
        im2.add(**self.intvls[1])
        obj2 = self._interval_metadata_constructor_(self.upper_bound, im2)

        self.assertReallyNotEqual(obj1, obj2)

    def test_copy_interval_metadata_empty(self):
        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)
        obj_copy = copy.copy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNone(obj_copy._interval_metadata)
        self.assertEqual(obj._interval_metadata, self.im)

    def test_copy_interval_metadata_none(self):
        obj = self._interval_metadata_constructor_(self.upper_bound)
        obj_copy = copy.copy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNone(obj._interval_metadata)
        self.assertIsNone(obj_copy._interval_metadata)

    def test_copy_interval_metadata(self):
        self.im.add(**self.intvls[1])
        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)
        obj_copy = copy.copy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNot(obj.interval_metadata, obj_copy.interval_metadata)
        self.assertIsNot(
            obj.interval_metadata._intervals, obj_copy.interval_metadata._intervals
        )
        for i, j in zip(
            obj.interval_metadata._intervals, obj_copy.interval_metadata._intervals
        ):
            self.assertIsNot(i, j)
            self.assertIsNot(i.metadata, j.metadata)
            for k in i.metadata:
                self.assertIs(i.metadata[k], j.metadata[k])

    def test_deepcopy_interval_metadata(self):
        self.im.add(**self.intvls[1])
        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)
        obj_copy = copy.deepcopy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNot(obj.interval_metadata, obj_copy.interval_metadata)
        self.assertIsNot(
            obj.interval_metadata._intervals, obj_copy.interval_metadata._intervals
        )
        for i, j in zip(
            obj.interval_metadata._intervals, obj_copy.interval_metadata._intervals
        ):
            self.assertIsNot(i, j)
            self.assertIsNot(i.metadata, j.metadata)
            self.assertIsNot(i.metadata["gene"], j.metadata["gene"])
            self.assertIs(i.metadata["product"], j.metadata["product"])

    def test_deepcopy_interval_metadata_empty(self):
        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)
        obj_copy = copy.deepcopy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNone(obj_copy._interval_metadata)
        self.assertEqual(obj._interval_metadata, self.im)

    def test_deepcopy_interval_metadata_none(self):
        obj = self._interval_metadata_constructor_(self.upper_bound, None)
        obj_copy = copy.deepcopy(obj)

        self.assertEqual(obj, obj_copy)
        self.assertIsNot(obj, obj_copy)

        self.assertIsNone(obj._interval_metadata)
        self.assertIsNone(obj_copy._interval_metadata)

    def test_deepcopy_memo_is_respected(self):
        # Basic test to ensure deepcopy's memo is passed through to recursive
        # deepcopy calls.
        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)
        memo = {}
        copy.deepcopy(obj, memo)
        self.assertGreater(len(memo), 1)

    def test_interval_metadata_getter(self):
        self.im.add(**self.intvls[0])
        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)
        self.assertIsInstance(obj.interval_metadata, IntervalMetadata)
        self.assertEqual(self.im, obj.interval_metadata)

        # Update existing metadata.
        obj.interval_metadata._intervals[0].metadata["gene"] = "sagB"
        self.assertNotEqual(obj.interval_metadata, self.im)
        self.im._intervals[0].metadata["gene"] = "sagB"
        self.assertEqual(obj.interval_metadata, self.im)

        # Add new interval feature.
        obj.interval_metadata.add(**self.intvls[1])
        self.im.add(**self.intvls[1])
        self.assertEqual(obj.interval_metadata, self.im)

    def test_interval_metadata_getter_no_interval_metadata(self):
        obj = self._interval_metadata_constructor_(self.upper_bound)
        self.assertIsNone(obj._interval_metadata)
        self.assertIsInstance(obj.interval_metadata, IntervalMetadata)
        self.assertEqual(obj.interval_metadata, self.im)
        self.assertIsNotNone(obj._interval_metadata)

    def test_interval_metadata_setter(self):
        obj = self._interval_metadata_constructor_(self.upper_bound)

        self.assertFalse(obj.has_interval_metadata())

        obj.interval_metadata = self.im
        self.assertFalse(obj.has_interval_metadata())
        self.assertEqual(obj.interval_metadata, self.im)

        self.im.add(**self.intvls[1])
        obj.interval_metadata = self.im
        self.assertTrue(obj.has_interval_metadata())
        self.assertEqual(obj.interval_metadata, self.im)

    def test_interval_metadata_setter_makes_copy(self):
        intvl = self.im.add(**self.intvls[1])
        obj = self._interval_metadata_constructor_(self.upper_bound)
        obj.interval_metadata = self.im

        self.assertEqual(obj.interval_metadata, self.im)
        self.assertIsNot(obj.interval_metadata, self.im)

        # Changing mutable value of metadata of the old interval
        # also changes obj.
        intvl.metadata["gene"].append("b")
        self.assertEqual(obj.interval_metadata, self.im)

        # Changing old interval doesn't change obj
        intvl.bounds = [(3, 6)]
        self.assertNotEqual(obj.interval_metadata, self.im)

    def test_interval_metadata_setter_len_mismatch(self):
        self.im.add(**self.intvls[1])
        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)

        for i in 0, 1, 3, 100:
            with self.assertRaisesRegex(
                ValueError, r"\(%d\).*\(%d\)" % (i, self.upper_bound)
            ):
                obj.interval_metadata = IntervalMetadata(i)

        self.assertEqual(obj.interval_metadata, self.im)

    def test_interval_metadata_setter_invalid_type(self):
        self.im.add(**self.intvls[0])
        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)

        for i in [2, None, "", {}, []]:
            with self.assertRaisesRegex(
                TypeError, "You must provide `IntervalMetadata` object"
            ):
                obj.interval_metadata = i

        self.assertEqual(self.im, obj.interval_metadata)

    def test_interval_metadata_setter_empty_upper_bound_is_none(self):
        im = IntervalMetadata(None)
        for i in [0, 1, 3, 100]:
            x = self._interval_metadata_constructor_(i)
            x.interval_metadata = im
            self.assertFalse(x.has_interval_metadata())
            # the upper bound is reset to seq/axis length
            self.assertEqual(x.interval_metadata.upper_bound, i)
            # original interval metadata upper bound is not changed
            self.assertIsNone(im.upper_bound)

    def test_interval_metadata_setter_upper_bound_is_none(self):
        im = IntervalMetadata(None)
        # populate im
        im.add(**self.intvls[0])
        im.add(**self.intvls[1])
        for i in [1000, 100]:
            x = self._interval_metadata_constructor_(i)
            x.interval_metadata = im
            # the upper bound is reset to seq/axis length
            self.assertEqual(x.interval_metadata.upper_bound, i)
            self.assertEqual(x.interval_metadata._intervals, im._intervals)
            # original interval metadata upper bound is not changed
            self.assertIsNone(im.upper_bound)

    def test_interval_metadata_setter_interval_bounds_larger_than_len(self):
        im = IntervalMetadata(None)
        # populate im
        im.add(**self.intvls[0])
        im.add(**self.intvls[1])
        for i in [0, 1, 3]:
            # error to reset upper bound to a smaller value than seq/axis len
            with self.assertRaisesRegex(
                ValueError, r"larger than upper bound \(%r\)" % i
            ):
                x = self._interval_metadata_constructor_(i)
                x.interval_metadata = im
            # original interval metadata upper bound is not changed
            self.assertIsNone(im.upper_bound)

    def test_interval_metadata_deleter_empty(self):
        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)

        del obj.interval_metadata
        self.assertIsNone(obj._interval_metadata)
        self.assertFalse(obj.has_interval_metadata())

        # Delete again. test idempotent
        del obj.interval_metadata
        self.assertIsNone(obj._interval_metadata)
        self.assertFalse(obj.has_interval_metadata())

    def test_interval_metadata_deleter(self):
        self.im.add(**self.intvls[0])
        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)

        del obj.interval_metadata
        self.assertIsNone(obj._interval_metadata)
        self.assertFalse(obj.has_interval_metadata())

    def test_has_interval_metadata(self):
        obj = self._interval_metadata_constructor_(self.upper_bound)
        self.assertFalse(obj.has_interval_metadata())

        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)
        self.assertFalse(obj.has_interval_metadata())

        self.im.add([(0, 1)])
        obj = self._interval_metadata_constructor_(self.upper_bound, self.im)
        self.assertTrue(obj.has_interval_metadata())



r"""Taxdump format (:mod:`skbio.io.format.taxdump`)
===============================================

.. currentmodule:: skbio.io.format.taxdump

The NCBI Taxonomy database dump (``taxdump``) format stores information of
organism names, classifications and other properties. It is a tabular format
with a delimiter: ``<tab><pipe><tab>`` between columns, and a line end
``<tab><pipe>`` after all columns. The file name usually ends with .dmp.

Format Support
--------------
**Has Sniffer: No**

+------+------+---------------------------------------------------------------+
|Reader|Writer|                          Object Class                         |
+======+======+===============================================================+
|Yes   |No    |:mod:`pandas.DataFrame`                                        |
+------+------+---------------------------------------------------------------+

Format Specification
--------------------

The NCBI taxonomy database [1]_ [2]_ hosts organism names and classifications.
It has a web portal [3]_ and an FTP download server [4]_. It is also accessible
using E-utilities [5]_. The database is being updated daily, and an archive is
generated every month. The data release has the file name ``taxdump``. It
consists of multiple .dmp files. These files serve different purposes, but they
follow a common format pattern:

- It is a tabular format.
- Column delimiter is ``<tab><pipe><tab>``.
- Line end is ``<tab><pipe>``.
- The first column is a numeric identifier, which usually represent taxa (i.e.,
  "TaxID"), but can also be genetic codes, citations or other entries.

The two most important files of the data release are ``nodes.dmp`` and
``names.dmp``. They store the hierarchical structure of the classification
system (i.e., taxonomy) and the names of organisms, respectively. They can be
used to construct the taxonomy tree of organisms.

The definition of columns of each .dmp file type are taken from [6]_ and [7]_.

``nodes.dmp``
^^^^^^^^^^^^^
+----------------+-------------------------------------+
|Name            |Description                          |
+================+=====================================+
|tax_id          |node id in GenBank taxonomy database |
+----------------+-------------------------------------+
|parent tax_id   |parent node id in GenBank taxonomy   |
|                |database                             |
+----------------+-------------------------------------+
|rank            |rank of this node (superkingdom,     |
|                |kingdom, ...)                        |
+----------------+-------------------------------------+
|embl code       |locus-name prefix; not unique        |
+----------------+-------------------------------------+
|division id     |see division.dmp file                |
+----------------+-------------------------------------+
|inherited div   |1 if node inherits division from     |
|flag (1 or 0)   |parent                               |
+----------------+-------------------------------------+
|genetic code id |see gencode.dmp file                 |
+----------------+-------------------------------------+
|inherited GC    |1 if node inherits genetic code from |
|flag (1 or 0)   |parent                               |
+----------------+-------------------------------------+
|mitochondrial   |see gencode.dmp file                 |
|genetic code id |                                     |
+----------------+-------------------------------------+
|inherited MGC   |1 if node inherits mitochondrial     |
|flag (1 or 0)   |gencode from parent                  |
+----------------+-------------------------------------+
|GenBank hidden  |1 if name is suppressed in GenBank   |
|flag (1 or 0)   |entry lineage                        |
+----------------+-------------------------------------+
|hidden subtree  |1 if this subtree has no sequence    |
|root flag       |data yet                             |
|(1 or 0)        |                                     |
+----------------+-------------------------------------+
|comments        |free-text comments and citations     |
+----------------+-------------------------------------+

Since 2018, NCBI releases "new taxonomy files" [8]_ (``new_taxdump``). The new
``nodes.dmp`` format is compatible with the classical format, plus five extra
columns after all aforementioned columns.

+----------------+-------------------------------------+
|Name            |Description                          |
+================+=====================================+
|plastid genetic |see gencode.dmp file                 |
|code id         |                                     |
+----------------+-------------------------------------+
|inherited PGC   |1 if node inherits plastid gencode   |
|flag (1 or 0)   |from parent                          |
+----------------+-------------------------------------+
|specified       |1 if species in the node's lineage   |
|species         |has formal name                      |
+----------------+-------------------------------------+
|hydrogenosome   |see gencode.dmp file                 |
|genetic code id |                                     |
+----------------+-------------------------------------+
|inherited HGC   |1 if node inherits hydrogenosome     |
|flag (1 or 0)   |gencode from parent                  |
+----------------+-------------------------------------+

``names.dmp``
^^^^^^^^^^^^^
+----------------+-------------------------------------+
|Name            |Description                          |
+================+=====================================+
|tax_id          |the id of node associated with this  |
|                |name                                 |
+----------------+-------------------------------------+
|name_txt        |name itself                          |
+----------------+-------------------------------------+
|unique name     |the unique variant of this name if   |
|                |name not unique                      |
+----------------+-------------------------------------+
|name class      |(synonym, common name, ...)          |
+----------------+-------------------------------------+

``division.dmp``
^^^^^^^^^^^^^^^^
+----------------+-------------------------------------+
|Name            |Description                          |
+================+=====================================+
|division id     |taxonomy database division id        |
+----------------+-------------------------------------+
|division cde    |GenBank division code (three         |
|                |characters)                          |
+----------------+-------------------------------------+
|division name   |e.g. BCT, PLN, VRT, MAM, PRI...      |
+----------------+-------------------------------------+
|comments        |                                     |
+----------------+-------------------------------------+

``gencode.dmp``
^^^^^^^^^^^^^^^
+----------------+-------------------------------------+
|Name            |Description                          |
+================+=====================================+
|genetic code id |GenBank genetic code id              |
+----------------+-------------------------------------+
|abbreviation    |genetic code name abbreviation       |
+----------------+-------------------------------------+
|name            |genetic code name                    |
+----------------+-------------------------------------+
|cde             |translation table for this genetic   |
|                |code                                 |
+----------------+-------------------------------------+
|starts          |start codons for this genetic code   |
+----------------+-------------------------------------+

Other types of .dmp files are currently not supported by scikit-bio. However,
the user may customize column definitions in using this utility. See below for
details.

Format Parameters
-----------------
The following format parameters are available in ``taxdump`` format:

- ``scheme``: The column definition scheme name of the input .dmp file.
  Available options are listed below. Alternatively, one can provide a custom
  scheme as defined in a name-to-data type dictionary.

  1. ``nodes``: The classical ``nodes.dmp`` scheme. It is also compatible with
     new ``nodes.dmp`` format, in which case only the columns defined by the
     classical format will be read.

  2. ``nodes_new``: The new ``nodes.dmp`` scheme.

  3. ``nodes_slim``: Only the first three columns: tax_id, parent_tax_id and
     rank, which are the minimum required information for constructing the
     taxonomy tree. It can be applied to both classical and new ``nodes.dmp``
     files. It can also handle custom files which only contains these three
     columns.

  4. ``names``: The ``names.dmp`` scheme.

  5. ``division``: The ``division.dmp`` scheme.

  6. ``gencode``: The ``gencode.dmp`` scheme.

.. note:: scikit-bio will read columns from leftmost till the number of columns
   defined in the scheme. Extra columns will be cropped.

Examples
--------
>>> from io import StringIO
>>> import skbio.io
>>> import pandas as pd
>>> fs = '\n'.join([
...     '1\t|\t1\t|\tno rank\t|',
...     '2\t|\t131567\t|\tsuperkingdom\t|',
...     '6\t|\t335928\t|\tgenus\t|'
... ])
>>> fh = StringIO(fs)

Read the file into a ``pd.DataFrame`` and specify that the "nodes_slim" scheme
should be used:

>>> df = skbio.io.read(fh, format="taxdump", into=pd.DataFrame,
...                    scheme="nodes_slim")
>>> df # doctest: +NORMALIZE_WHITESPACE
        parent_tax_id          rank
tax_id
1                   1       no rank
2              131567  superkingdom
6              335928         genus

References
----------
.. [1] Federhen, S. (2012). The NCBI taxonomy database. Nucleic acids
   research, 40(D1), D136-D143.
.. [2] Schoch, C. L., Ciufo, S., Domrachev, M., Hotton, C. L., Kannan, S.,
   Khovanskaya, R., ... & Karsch-Mizrachi, I. (2020). NCBI Taxonomy: a
   comprehensive update on curation, resources and tools. Database, 2020.
.. [3] https://www.ncbi.nlm.nih.gov/taxonomy
.. [4] https://ftp.ncbi.nlm.nih.gov/pub/taxonomy/
.. [5] Kans, J. (2022). Entrez direct: E-utilities on the UNIX command line.
   In Entrez Programming Utilities Help [Internet]. National Center for
   Biotechnology Information (US).
.. [6] https://ftp.ncbi.nlm.nih.gov/pub/taxonomy/taxdump_readme.txt
.. [7] https://ftp.ncbi.nlm.nih.gov/pub/taxonomy/new_taxdump/taxdump_readme.txt
.. [8] https://ncbiinsights.ncbi.nlm.nih.gov/2018/02/22/new-taxonomy-files-
       available-with-lineage-type-and-host-information/


"""  # noqa: D205, D415

# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

import pandas as pd

from skbio.io import create_format


taxdump = create_format("taxdump")

_taxdump_column_schemes = {
    "nodes_slim": {"tax_id": int, "parent_tax_id": int, "rank": str},
    "nodes": {
        "tax_id": int,
        "parent_tax_id": int,
        "rank": str,
        "embl_code": str,
        "division_id": int,
        "inherited_div_flag": bool,
        "genetic_code_id": int,
        "inherited_GC_flag": bool,
        "mitochondrial_genetic_code_id": int,
        "inherited_MGC_flag": bool,
        "GenBank_hidden_flag": bool,
        "hidden_subtree_root_flag": bool,
        "comments": str,
    },
    "names": {"tax_id": int, "name_txt": str, "unique_name": str, "name_class": str},
    "division": {
        "division_id": int,
        "division_cde": str,
        "division_name": str,
        "comments": str,
    },
    "gencode": {
        "genetic_code_id": int,
        "abbreviation": str,
        "name": str,
        "cde": str,
        "starts": str,
    },
}

_taxdump_column_schemes["nodes_new"] = dict(
    _taxdump_column_schemes["nodes"],
    **{
        "plastid_genetic_code_id": bool,
        "inherited_PGC_flag": bool,
        "specified_species": bool,
        "hydrogenosome_genetic_code_id": int,
        "inherited_HGC_flag": bool,
    },
)


@taxdump.reader(pd.DataFrame, monkey_patch=False)
def _taxdump_to_data_frame(fh, scheme):
    """Read a taxdump file into a data frame.

    Parameters
    ----------
    fh : file handle
        Input taxdump file
    scheme : str
        Name of column scheme

    Returns
    -------
    pd.DataFrame
        Parsed table

    """
    if isinstance(scheme, str):
        if scheme not in _taxdump_column_schemes:
            raise ValueError(f'Invalid taxdump column scheme: "{scheme}".')
        scheme = _taxdump_column_schemes[scheme]
    names = list(scheme.keys())
    try:
        return pd.read_csv(
            fh,
            sep="\t\\|(?:\t|$)",
            engine="python",
            index_col=0,
            names=names,
            dtype=scheme,
            usecols=range(len(names)),
        )
    except ValueError:
        raise ValueError("Invalid taxdump file format.")



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

import numpy as np
import pandas as pd
from scipy.linalg import svd

from skbio.sequence import Sequence
from skbio._base import SkbioObject
from skbio.stats.ordination import OrdinationResults
from skbio.diversity import beta_diversity


def _repr_helper(rstr, org_name, new_name, dim_name, regex_match, shape):
    rstr = rstr.replace(org_name, new_name)
    n_indent = 4  # see Sequence.__repr__
    indent = " " * n_indent
    rstr = rstr.replace(
        regex_match,
        f"{dim_name} dimension: {shape}\n{indent}has gaps",
    )
    return rstr


class Embedding(SkbioObject):
    r"""Embedding for a biological object.

    Parameters
    ----------
    embedding : array_like
        Embedding matrix where the first axis is indexed by `ids`.
    ids : array_like
        IDs of biological objects.

    """

    @property
    def embedding(self):
        r"""The embedding tensor."""
        return self._embedding

    @property
    def ids(self):
        r"""IDs corresponding to each row of the embedding."""
        # each embedding row corresponds to an id
        return self._ids

    def __init__(self, embedding, ids, **kwargs):
        # make sure that the embedding has the same length as the sequence
        ids_len = len(ids)
        if embedding.shape[0] != ids_len:
            raise ValueError(
                f"The embedding ({embedding.shape[0]}) must have the "
                f"same length as the ids ({ids_len})."
            )

        self._embedding = np.asarray(embedding)
        self._ids = np.asarray(ids)

    def __str__(self):
        raise NotImplementedError("This method should be implemented by subclasses.")

    def bytes(self):
        r"""Bytes representation of string encoding."""
        seq = np.frombuffer(str(self).encode("ascii"), dtype=np.uint8)
        return seq


class SequenceEmbedding(Embedding):
    r"""Embedding for a biological sequence.

    Parameters
    ----------
    embedding : array_like
        The embedding of the sequence. Row vectors correspond to the latent character
        coordinates.
    sequence : str, Sequence, or 1D ndarray
        Characters representing the sequence itself.

    See Also
    --------
    Embedding
    skbio.sequence.Sequence

    """

    def __init__(self, embedding, sequence, **kwargs):
        if isinstance(sequence, Sequence):
            sequence = str(sequence)
        if isinstance(sequence, str):
            sequence = sequence.encode("ascii")
        seq = np.frombuffer(sequence, dtype=np.uint8)
        super(SequenceEmbedding, self).__init__(embedding, seq, **kwargs)

    def __str__(self):
        r"""String representation of the underlying sequence."""
        return str(self._ids.tobytes().decode("ascii"))

    @property
    def sequence(self):
        r"""String representation of the underlying sequence."""
        return str(self)

    def __repr__(self):
        r"""Return a string representation of the SequenceEmbedding object.

        Returns
        -------
        str
            String representation of the SequenceEmbedding object.

        See Also
        --------
        skbio.sequence.Protein

        """
        seq = Sequence(self.sequence)
        rstr = _repr_helper(
            repr(seq),
            "Sequence",
            "SequenceEmbedding",
            "embedding",
            regex_match="length",
            shape=self.embedding.shape[1],
        )
        return rstr


class EmbeddingVector(Embedding):
    r"""Vector representation for a biological entity.

    Parameters
    ----------
    vector : 1D or 2D array_like
        The vector representation of the sequence. Typically a 1D array. Can also be a
        2D array with only one row.
    sequence : str, Sequence, or 1D ndarray
        Characters representing the sequence itself.

    See Also
    --------
    Embedding

    """

    def __init__(self, vector, obj, **kwargs):
        super(EmbeddingVector, self).__init__(vector, obj, **kwargs)

    def __str__(self):
        return self._ids[0].decode("ascii")

    @property
    def vector(self):
        r"""Vector representation for the biological entity."""
        return self._embedding.squeeze()

    @property
    def embedding(self):
        r"""The embedding tensor."""
        return self._embedding.reshape(1, -1)


class SequenceVector(EmbeddingVector):
    r"""Vector representation for a biological sequence.

    Parameters
    ----------
    vector : 1D or 2D array_like
        The vector representation of the sequence. Typically a 1D array. Can also be a
        2D array with only one row.
    sequence : str, Sequence, or 1D ndarray
        Characters representing the sequence itself.

    See Also
    --------
    EmbeddingVector
    skbio.sequence.Sequence

    """

    def __init__(self, vector, sequence, **kwargs):
        vector = np.atleast_2d(vector)
        if vector.shape[0] != 1:
            raise ValueError("Only one vector per sequence is allowed.")

        if isinstance(sequence, Sequence):
            sequence = str(sequence)
        if isinstance(sequence, str):
            sequence = sequence.encode("ascii")
        sequence = np.array([sequence], dtype="O")

        super(SequenceVector, self).__init__(vector, sequence, **kwargs)

    @property
    def sequence(self):
        r"""String representation of the underlying sequence."""
        return str(self)

    def __repr__(self):
        r"""Return a string representation of the SequenceVector object.

        Returns
        -------
        str
            A string representation of the SequenceVector object.

        See Also
        --------
        skbio.sequence.Sequence

        """
        seq = Sequence(str(self))
        rstr = _repr_helper(
            repr(seq),
            "Sequence",
            "SequenceVector",
            "vector",
            regex_match="length",
            shape=self.embedding.shape[1],
        )
        return rstr


def embed_vec_to_numpy(vectors, validate=True):
    r"""Convert an iterable of EmbeddingVector objects to a NumPy array.

    Parameters
    ----------
    vectors : iterable of EmbeddingVector objects
        An iterable of EmbeddingVector objects, or objects that
        subclass EmbeddingVector.
    validate : bool, optional
        If ``True``, validate that all vectors have the same length
        and are valid types.

    Returns
    -------
    ndarray of shape (n_objects, n_features)
        A NumPy array where n_features corresponds to the dimensionality of the latent
        space.

    Raises
    ------
    ValueError
        If the vectors do not have the same length.

    """
    if validate:
        subcls = [issubclass(type(ev), EmbeddingVector) for ev in vectors]
        if not all(subcls):
            raise ValueError(
                "Input iterable contains objects that "
                "do not subclass EmbeddingVector."
            )

        types = [type(ev) for ev in vectors]
        if not all(t == types[0] for t in types):
            raise ValueError("All objects must be of the same type.")

        lens = [len(ev.vector) for ev in vectors]
        if not all(ln == lens[0] for ln in lens):
            raise ValueError("All vectors must have the same length.")

    data = np.vstack([ev.vector for ev in vectors])
    return data


def embed_vec_to_distances(vectors, metric="euclidean", validate=True):
    r"""Convert EmbeddingVector objects to a DistanceMatrix object.

    Parameters
    ----------
    vectors : iterable of EmbeddingVector objects
        An iterable of EmbeddingVector objects, or objects that
        subclass EmbeddingVector.
    metric : str or callable, optional
        The distance metric to use. Must be a valid metric for
        ``scipy.spatial.distance.pdist``.
    validate : bool, optional
        If ``True``, validate that all vectors have the same length
        and are valid types.

    Returns
    -------
    DistanceMatrix
        A distance matrix representing pairwise distances among objects calculated by
        the given metric.

    See Also
    --------
    skbio.stats.distance.DistanceMatrix

    """
    data = embed_vec_to_numpy(vectors, validate=validate)
    ids = [str(ev) for ev in vectors]
    return beta_diversity(metric, data, ids)


def embed_vec_to_ordination(vectors, validate=True):
    r"""Convert EmbeddingVector objects to an Ordination object.

    A singular value decomposition (SVD) is performed on the data.

    Parameters
    ----------
    vectors : iterable of EmbeddingVector objects
        An iterable of EmbeddingVector objects, or objects that subclass
        EmbeddingVector.
    validate : bool, optional
        If ``True``, validate that all vectors have the same length and are valid
        types.

    Returns
    -------
    OrdinationResults
        Ordination results with objects as samples and latent variables as features.

    See Also
    --------
    skbio.stats.ordination.OrdinationResults

    """
    data = embed_vec_to_numpy(vectors, validate=validate)
    u, s, vh = svd(data, full_matrices=False)
    eigvals = s**2
    short_name = "SVD"
    long_name = "Singular Value Decomposition"
    # note that we are moving half of the singular values
    # in the eigvals to the samples and the other half to the features
    # this is to help with the interpretation of the ordination
    # if visualizing with biplots
    ordr = OrdinationResults(
        short_method_name=short_name,
        long_method_name=long_name,
        eigvals=eigvals,
        proportion_explained=eigvals / eigvals.sum(),
        samples=pd.DataFrame(u @ np.diag(s), index=[str(ev) for ev in vectors]),
        features=pd.DataFrame(vh.T, index=range(data.shape[1])),
    )
    return ordr


def embed_vec_to_dataframe(vectors, validate=True):
    r"""Convert a list of SequenceVector objects to a pandas DataFrame.

    Parameters
    ----------
    vectors : iterable of EmbeddingVector objects
        An iterable of EmbeddingVector objects, or objects that
        subclass EmbeddingVector.
    validate : bool, optional
        If ``True``, validate that all vectors have the same length
        and are valid types.

    Returns
    -------
    pd.DataFrame
        Data frame containing the embedding vectors as rows (index) and object IDs as
        columns.

    See Also
    --------
    pd.DataFrame

    """
    data = embed_vec_to_numpy(vectors, validate=validate)
    return pd.DataFrame(data, index=[str(ev) for ev in vectors])



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

import re

from skbio.metadata import IntervalMetadata
from skbio.io.format._base import _line_generator
from skbio.io import FileFormatError


def _vocabulary_change(format="insdc", read_in=True):
    """Return a dict that converts between memory and output vocabulary."""
    convert = {
        "phase": {"insdc": "codon_start"},
        "source": {"insdc": "inference"},
        "db_xref": {"gff3": "Dbxref"},
        "note": {"gff3": "Note"},
    }
    if read_in:
        return {v[format]: k for k, v in convert.items() if format in v}
    else:
        return {k: v[format] for k, v in convert.items() if format in v}


def _vocabulary_skip(format="insdc"):
    """Return vocabluaries skipped for auto disk output, given a format.

    Return a list of vocabularies that should be skipped when auto output to disk
    for the specified format.

    """
    skip = {
        "type": ("insdc", "gff3"),
        "ID": ("insdc"),
        "translation": ("gff3"),
        "strand": ("insdc"),
    }
    return [k for k, v in skip.items() if format in v]


def _yield_section(is_another_section, **kwargs):
    """Return function that returns successive sections from file.

    Parameters
    ----------
    is_another_section : callable
        It takes a string as input and return a boolean indicating
        a new section starts.
    kwargs : dict, optional
        Keyword arguments will be passed to `_line_generator`.

    Returns
    -------
    function
        A function accept a list of lines as input and return
        a generator to yield section one by one.

    """

    def parser(lines):
        curr = []
        for line in _line_generator(lines, **kwargs):
            # if we find another, return the previous section
            if is_another_section(line):
                if curr:
                    yield curr
                    curr = []
            curr.append(line)
        # don't forget to return the last section in the file
        if curr:
            yield curr

    return parser


def _parse_section_default(
    lines, label_delimiter=None, join_delimiter=" ", return_label=False
):
    """Parse sections in default way.

    Do 2 things:
        1. split first line with label_delimiter for label
        2. join all the lines into one str with join_delimiter.
    """
    data = []
    label = None
    line = lines[0]

    items = line.split(label_delimiter, 1)

    if len(items) == 2:
        label, section = items
    else:
        label = items[0]
        section = ""
    data.append(section)

    data.extend(lines[1:])
    data = join_delimiter.join(i.strip() for i in data)
    if return_label:
        return label, data
    else:
        return data


def _serialize_section_default(header, obj, indent=12):
    return "{header:<{indent}}{obj}\n".format(header=header, obj=obj, indent=indent)


def _parse_feature_table(lines, length):
    """Parse DDBJ/ENA/GenBank Feature Table."""
    imd = IntervalMetadata(length)
    # skip the 1st FEATURES line
    if lines[0].startswith("FEATURES"):
        lines = lines[1:]
    # magic number 21: the lines following header of each feature
    # are indented with 21 spaces.
    feature_indent = " " * 21
    section_splitter = _yield_section(
        lambda x: not x.startswith(feature_indent), skip_blanks=True, strip=False
    )

    for section in section_splitter(lines):
        _parse_single_feature(section, imd)
    return imd


def _parse_single_feature(lines, imd):
    """Parse a feature and add it to ``IntervalMetadata`` object.

    Parameters
    ----------
    imd : IntervalMetadata
        An IntervalMetadata object to which the parsed feature will be added.
    lines : list of strings
        A list of strings representing the lines of text to be parsed.

    """
    voca_change = _vocabulary_change("insdc")

    # each component of a feature starts with '/', except the 1st
    # component of location.
    section_splitter = _yield_section(lambda x: x.startswith("/"), strip=True)
    section_iter = section_splitter(lines)

    # 1st section is location
    section = next(section_iter)
    feature_type, feature_loc = _parse_section_default(
        section, join_delimiter="", return_label=True
    )

    metadata = {"type": feature_type, "__location": feature_loc}

    intvl = imd.add(*_parse_loc_str(feature_loc))

    for section in section_iter:
        # following sections are Qualifiers
        k, v = _parse_section_default(
            section, label_delimiter="=", join_delimiter=" ", return_label=True
        )
        # 1st char is '/'
        k = k[1:]
        if k in voca_change:
            k = voca_change[k]

        if k == "phase":
            v = int(v) - 1

        # some Qualifiers can appear multiple times
        if k in metadata:
            if not isinstance(metadata[k], list):
                metadata[k] = [metadata[k]]
            metadata[k].append(v)
        else:
            metadata[k] = v

    intvl.metadata.update(metadata)


def _parse_loc_str(loc_str):
    """Parse location string.

    .. warning: This converts coordinates to 0-based from 1-based
    GenBank coordinate system.

    The location descriptor can be one of the following [1]_:
    (a) a single base number. e.g. 467
    (b) a site between two indicated adjoining bases. e.g. 123^124
    (c) a single base chosen from within a specified range of bases (not
        allowed for new entries). e.g. 102.110
    (d) the base numbers delimiting a sequence span. e.g.340..565
    (e) a remote entry identifier followed by a local location
        descriptor (i.e., a-d). e.g. J00194.1:100..202

    Notes
    -----
    This does not fully handle (e) case. It will discard the remote
    entry part and only keep the local part. When it parses locations
    across strand (e.g. "complement(123..145),200..209"), it will
    record all the span parts but will record strand as negative.

    References
    ----------
    .. [1] http://www.insdc.org/files/feature_table.html#3.4

    """
    # define the tokens
    operators = ["join", "complement", "order"]
    LPAREN = r"(?P<LPAREN>\()"
    RPAREN = r"(?P<RPAREN>\))"
    COMMA = r"(?P<COMMA>,)"
    WS = r"(?P<WS>\s+)"
    a = r"(?P<A>\d+)"
    b = r"(?P<B>\d+\^\d+)"
    c = r"(?P<C>\d+\.\d+)"
    d = r"(?P<D><?\d+\.\.>?\d+)"
    e_left = r"(?P<EL><?[a-zA-Z_0-9\.]+:\d+\.\.>?\d+)"
    e_right = r"(?P<ER><?\d+\.\.>?[a-zA-Z_0-9\.]+:\d+)"
    illegal = r"(?P<ILLEGAL>.+)"
    # The order of tokens in the master regular expression also
    # matters. When matching, re tries to match pattens in the order
    # specified. Thus, if a pattern happens to be a substring of a
    # longer pattern, you need to make sure the longer pattern goes
    # first.
    master_pat = re.compile(
        "|".join(
            operators
            + [WS, LPAREN, RPAREN, COMMA, b, c, d, e_left, e_right, a, illegal]
        )
    )

    scanner = master_pat.scanner(loc_str)

    bounds = []
    fuzzy = []

    metadata = {"strand": "+"}

    for m in iter(scanner.match, None):
        p, v = m.lastgroup, m.group()
        if v == "complement":
            metadata["strand"] = "-"
        elif p == "A":
            start = int(v)
            bounds.append((start - 1, start))
            fuzzy.append((False, False))
        elif p == "B":
            start, end = v.split("^")
            start = int(start)
            bounds.append((start - 1, start))
            fuzzy.append((False, False))
        elif p == "C" or p == "D":
            if p == "C":
                start, end = v.split(".")
            else:
                start, end = v.split("..")
            fuzzy_s = fuzzy_e = False
            if start.startswith("<"):
                start = start[1:]
                fuzzy_s = True
            if end.startswith(">"):
                end = end[1:]
                fuzzy_e = True
            bounds.append((int(start) - 1, int(end)))
            fuzzy.append((fuzzy_s, fuzzy_e))
        elif p == "ILLEGAL":
            raise FileFormatError('Could not parse location string: "%s"' % loc_str)

    return bounds, fuzzy, metadata


def _serialize_feature_table(intervals, indent=21):
    """Serialize a list of intervals into a feature table format.

    Parameters
    ----------
    intervals : list of ``Interval``
        A list of Interval objects representing the intervals to be serialized.
    indent : int, optional
        The number of spaces to indent each serialized feature. Defaults to 21.

    """
    for intvl in intervals:
        yield _serialize_single_feature(intvl, indent)


def _serialize_single_feature(intvl, indent=21):
    """Serialize a single interval into feature format.

    Parameters
    ----------
    intvl : Interval
        The Interval object representing the interval to be serialized.
    indent : int, optional
        The number of spaces to indent each serialized feature. Defaults to 21.

    """
    # there are 5 spaces before Feature Key starts.
    padding = " " * 5
    qualifiers = []
    md = intvl.metadata
    voca_skip = _vocabulary_skip("insdc")
    voca_change = _vocabulary_change("insdc", read_in=False)
    # sort it so the output order is deterministic
    for k in sorted(md):
        if k.startswith("__") or k in voca_skip:
            continue
        v = md[k]
        if k == "phase":
            v = str(v + 1)
        if k in voca_change:
            k = voca_change[k]
        if isinstance(v, list):
            for vi in v:
                qualifiers.append(_serialize_qualifier(k, vi))
        else:
            qualifiers.append(_serialize_qualifier(k, v))

    if "__location" in md:
        loc = md["__location"]
    else:
        loc = _serialize_location(intvl)
    # the qualifiers start at column 22
    qualifiers = [" " * indent + i for i in qualifiers]
    return "{header:<{indent}}{loc}\n{qualifiers}\n".format(
        header=padding + md["type"],
        loc=loc,
        indent=indent,
        qualifiers="\n".join(qualifiers),
    )


def _serialize_location(intvl):
    loc = []
    for bound, fuzzy in zip(intvl.bounds, intvl.fuzzy):
        start, end = bound
        start += 1
        if start == end:
            s = str(start)
        elif fuzzy[0] and fuzzy[1]:
            s = "<%d..>%d" % (start, end)
        elif fuzzy[0] and not fuzzy[1]:
            s = "<%d..%d" % (start, end)
        elif not fuzzy[0] and fuzzy[1]:
            s = "%d..>%d" % (start, end)
        else:
            s = "%d..%d" % (start, end)
        loc.append(s)
    if len(loc) > 1:
        loc_str = "join({})".format(",".join(loc))
    else:
        loc_str = loc[0]
    if intvl.metadata.get("strand") == "-":
        loc_str = "complement({})".format(loc_str)
    return loc_str


def _serialize_qualifier(key, value):
    """Serialize a Qualifier in a feature.

    Parameters
    ----------
    key : str
        The key of the Qualifier, representing the type or name of the information.
    value : int, str
        The value associated with the Qualifier.

    """
    # if value is empty
    if not value:
        return "/%s" % key

    return "/{k}={v}".format(k=key, v=value)



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------


from skbio.metadata._repr import _MetadataReprBuilder


class _TabularMSAReprBuilder(_MetadataReprBuilder):
    def __init__(self, msa, width, indent):
        super(_TabularMSAReprBuilder, self).__init__(msa, width, indent)
        self._ellipse_insert = " ... "

    def _process_header(self):
        cls_name = self._obj.__class__.__name__
        if self._obj.dtype is not None:
            dtype_class = "[" + self._obj.dtype.__name__ + "]"
        else:
            dtype_class = ""
        self._lines.add_line(cls_name + dtype_class)
        self._lines.add_separator()

    def _process_data(self):
        num_sequences = self._obj.shape.sequence
        num_positions = self._obj.shape.position

        # catch case of all empty sequences
        if num_positions > 0:
            # display all sequences if we can, else display the first two and
            # last two sequences separated by ellipsis
            if num_sequences <= 5:
                self._lines.add_lines(self._format_sequences(range(num_sequences)))
            else:
                self._lines.add_lines(self._format_sequences(range(2)))
                self._lines.add_line("...")
                self._lines.add_lines(
                    self._format_sequences(range(num_sequences - 2, num_sequences))
                )

    def _format_sequences(self, sequence_indices):
        lines = []
        for line_index in sequence_indices:
            seq_str = str(self._obj._get_sequence_iloc_(line_index))
            if len(seq_str) <= self._width:
                formatted_seq = seq_str
            else:
                formatted_seq = (
                    seq_str[0 : self._num_characters_before_ellipse()]
                    + self._ellipse_insert
                    + seq_str[-self._num_characters_after_ellipse() :]
                )
            lines.append(formatted_seq)
        return lines

    def _num_characters_before_ellipse(self):
        return int(self._num_characters_to_display() / 2)

    def _num_characters_after_ellipse(self):
        return self._num_characters_to_display() - self._num_characters_before_ellipse()

    def _num_characters_to_display(self):
        return self._width - len(self._ellipse_insert)



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

import unittest

from skbio import Sequence, DNA, RNA, Protein
from skbio import read
from skbio.util import get_data_path
from skbio.io import QSeqFormatError
from skbio.io.format.qseq import _qseq_to_generator, _qseq_sniffer
import numpy as np


def _drop_kwargs(kwargs, *args):
    for arg in args:
        if arg in kwargs:
            kwargs.pop(arg)


class TestQSeqBase(unittest.TestCase):
    def setUp(self):
        self.valid_files = [
            (get_data_path('qseq_single_seq_sanger'), [
                {'variant': 'sanger'},
                {'phred_offset': 33},
            ], [
                {'id': 'sanger_1:3:34:-30:30#0/2',
                 'sequence': 'ACGTACGTACGTACGTACGTACGTACTTTTTTTTTTACGTACGTACG'
                             'TACGTACGTACGTACGTACGTACGTACGTACGTACGTACGTACGTAC',
                 'quality': [26, 26, 29, 31, 33, 34, 36, 37, 38, 39, 41, 42,
                             43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54,
                             55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66,
                             67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78,
                             79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90,
                             91, 92, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93,
                             93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 93,
                             93, 93, 93, 93, 93, 93, 93, 93, 93, 93],
                 'machine_name': 'sanger',
                 'run_number': 1,
                 'lane_number': 3,
                 'tile_number': 34,
                 'x': -30,
                 'y': 30,
                 'index': 0,
                 'read_number': 2}
            ]),

            (get_data_path('qseq_multi_seq_illumina1.3'), [
                {'variant': 'illumina1.3'},
                {'phred_offset': 64}
            ], [
                {'id': 'illumina_1:3:34:-30:30#0/1',
                 'sequence': 'ACG....ACGTAC',
                 'quality': [50, 53, 2, 2, 2, 2, 50, 2, 3, 5, 6, 7, 8],
                 'machine_name': 'illumina',
                 'run_number': 1,
                 'lane_number': 3,
                 'tile_number': 34,
                 'x': -30,
                 'y': 30,
                 'index': 0,
                 'read_number': 1},
                {'id': 'illumina_1:3:35:-30:30#0/2',
                 'sequence': 'ACGTA.AATAAAC',
                 'quality': [39, 37, 20, 33, 1, 33, 38, 40, 55, 49, 1, 1, 38],
                 'machine_name': 'illumina',
                 'run_number': 1,
                 'lane_number': 3,
                 'tile_number': 35,
                 'x': -30,
                 'y': 30,
                 'index': 0,
                 'read_number': 2}
            ]),

            (get_data_path('qseq_multi_seq_illumina1.3'), [
                {'variant': 'illumina1.3', 'filter': False, 'seq_num': 1},
                {'phred_offset': 64, 'filter': False, 'seq_num': 2},
                {'variant': 'illumina1.3', 'filter': False, 'seq_num': 3,
                 'constructor': Protein},
                {'phred_offset': 64, 'filter': False, 'seq_num': 4,
                 'constructor': DNA},
            ], [
                {'id': 'illumina_1:3:34:-30:30#0/1',
                 'sequence': 'ACG....ACGTAC',
                 'quality': [50, 53, 2, 2, 2, 2, 50, 2, 3, 5, 6, 7, 8],
                 'machine_name': 'illumina',
                 'run_number': 1,
                 'lane_number': 3,
                 'tile_number': 34,
                 'x': -30,
                 'y': 30,
                 'index': 0,
                 'read_number': 1},
                {'id': 'illumina_1:3:34:30:-30#0/1',
                 'sequence': 'CGGGCATTGCA',
                 'quality': [3, 7, 7, 7, 3, 33, 51, 36, 7, 3, 1],
                 'machine_name': 'illumina',
                 'run_number': 1,
                 'lane_number': 3,
                 'tile_number': 34,
                 'x': 30,
                 'y': -30,
                 'index': 0,
                 'read_number': 1},
                {'id': 'illumina_1:3:35:-30:30#0/2',
                 'sequence': 'ACGTA.AATAAAC',
                 'quality': [39, 37, 20, 33, 1, 33, 38, 40, 55, 49, 1, 1, 38],
                 'machine_name': 'illumina',
                 'run_number': 1,
                 'lane_number': 3,
                 'tile_number': 35,
                 'x': -30,
                 'y': 30,
                 'index': 0,
                 'read_number': 2},
                {'id': 'illumina_1:3:35:30:-30#0/3',
                 'sequence': 'CATTTAGGA.TGCA',
                 'quality': [52, 42, 38, 44, 43, 1, 6, 46, 43, 11, 39, 40, 54,
                             13],
                 'machine_name': 'illumina',
                 'run_number': 1,
                 'lane_number': 3,
                 'tile_number': 35,
                 'x': 30,
                 'y': -30,
                 'index': 0,
                 'read_number': 3}
            ])
        ]

        self.invalid_files = [
            (get_data_path('whitespace_only'), [
                {},
                {'variant': 'sanger'}
            ], [
                'blank line',
            ], QSeqFormatError),

            (get_data_path('tsv_10_fields'), [
                {},
                {'variant': 'sanger'},
                {'variant': 'solexa'}
            ], [
                'read',
                '[1, 3]'
            ], QSeqFormatError),

            (get_data_path('tsv_8_fields'), [
                {},
                {'variant': 'sanger'},
                {'variant': 'solexa'}
            ], [
                '8',
                '10 or 11'
            ], QSeqFormatError),


            (get_data_path('qseq_invalid_filter'), [
                {},
                {'phred_offset': 33},
                {'variant': 'solexa'},
                {'variant': 'illumina1.3'},
                {'variant': 'illumina1.8'}
            ], [
                'filter',
                '0 or 1',
            ], QSeqFormatError),

            (get_data_path('qseq_invalid_read'), [
                {},
                {'phred_offset': 33},
                {'variant': 'solexa'},
                {'variant': 'illumina1.3'},
                {'variant': 'illumina1.8'}
            ], [
                'read',
                '[1, 3]',
            ], QSeqFormatError),

            (get_data_path('qseq_invalid_x'), [
                {},
                {'phred_offset': 33},
                {'variant': 'solexa'},
                {'variant': 'illumina1.3'},
                {'variant': 'illumina1.8'}
            ], [
                'x',
                'integer',
            ], QSeqFormatError),

            (get_data_path('qseq_invalid_y'), [
                {},
                {'phred_offset': 33},
                {'variant': 'solexa'},
                {'variant': 'illumina1.3'},
                {'variant': 'illumina1.8'}
            ], [
                'y',
                'integer',
            ], QSeqFormatError),

            (get_data_path('qseq_invalid_lane'), [
                {},
                {'phred_offset': 33},
                {'variant': 'solexa'},
                {'variant': 'illumina1.3'},
                {'variant': 'illumina1.8'}
            ], [
                'lane',
                'positive integer',
            ], QSeqFormatError),

            (get_data_path('qseq_invalid_tile'), [
                {},
                {'phred_offset': 33},
                {'variant': 'solexa'},
                {'variant': 'illumina1.3'},
                {'variant': 'illumina1.8'}
            ], [
                'tile',
                'positive integer',
            ], QSeqFormatError)
        ]


class TestQSeqToGenerator(TestQSeqBase):

    def setUp(self):
        super(TestQSeqToGenerator, self).setUp()
        self.valid_files += [
            (get_data_path('empty'), [{}, {'variant': 'sanger'}], [])
        ]

        self.invalid_files += [
            (get_data_path('qseq_single_seq_sanger'), [
                {'variant': 'illumina1.3'},
                {'variant': 'illumina1.8'}
            ], [
                'out of range',
                '[0, 62]'
            ], ValueError)
        ]

    def test_invalid_files(self):
        for invalid, kwargs, errors, etype in self.invalid_files:
            with self.assertRaises(etype) as cm:
                for kwarg in kwargs:
                    _drop_kwargs(kwarg, 'seq_num')
                    list(_qseq_to_generator(invalid, **kwarg))
            for e in errors:
                self.assertIn(e, str(cm.exception))

    def test_valid_files(self):
        for valid, kwargs, components in self.valid_files:
            for kwarg in kwargs:
                _drop_kwargs(kwarg, 'seq_num')
                constructor = kwarg.get('constructor', Sequence)
                expected = [
                    constructor(
                        c['sequence'],
                        metadata={'id': c['id'],
                                  'machine_name': c['machine_name'],
                                  'run_number': c['run_number'],
                                  'lane_number': c['lane_number'],
                                  'tile_number': c['tile_number'],
                                  'x': c['x'],
                                  'y': c['y'],
                                  'index': c['index'],
                                  'read_number': c['read_number']},
                        positional_metadata={
                            'quality': np.array(c['quality'], dtype=np.uint8)})
                    for c in components]

                observed = list(_qseq_to_generator(valid, **kwarg))
                self.assertEqual(len(expected), len(observed))
                for o, e in zip(observed, expected):
                    self.assertEqual(o, e)


class TestQSeqToSequences(TestQSeqBase):
    def test_invalid_files(self):
        for constructor in [Sequence, DNA, RNA, Protein]:
            for invalid, kwargs, errors, etype in self.invalid_files:
                with self.assertRaises(etype) as cm:
                    for kwarg in kwargs:
                        _drop_kwargs(kwarg, 'constructor', 'filter')

                        read(invalid, format='qseq', verify=False,
                             into=constructor, **kwarg)
                for e in errors:
                    self.assertIn(e, str(cm.exception))

    def test_valid_files(self):
        for constructor in [Sequence, DNA, RNA, Protein]:
            for valid, kwargs, components in self.valid_files:
                for observed_kwargs in kwargs:
                    expected_kwargs = {}
                    # Currently not validating the alphabet for qseq
                    # files that are read in for this test.
                    if hasattr(constructor, 'alphabet'):
                        observed_kwargs['validate'] = False
                        expected_kwargs['validate'] = False
                    _drop_kwargs(observed_kwargs, 'constructor', 'filter')

                    seq_num = observed_kwargs.get('seq_num', 1)
                    c = components[seq_num - 1]
                    expected = constructor(
                        c['sequence'],
                        metadata={'id': c['id'],
                                  'machine_name': c['machine_name'],
                                  'run_number': c['run_number'],
                                  'lane_number': c['lane_number'],
                                  'tile_number': c['tile_number'],
                                  'x': c['x'],
                                  'y': c['y'],
                                  'index': c['index'],
                                  'read_number': c['read_number']},
                        positional_metadata={
                            'quality': np.array(c['quality'], np.uint8)},
                        **expected_kwargs)

                    observed = read(valid, into=constructor,
                                    format='qseq', verify=False,
                                    **observed_kwargs)
                    self.assertEqual(observed, expected)


class TestQSeqSniffer(TestQSeqBase):

    def setUp(self):
        super(TestQSeqSniffer, self).setUp()
        self.invalid_files += [
            (get_data_path('empty'), None, None, None)
        ]

    def test_qseq_sniffer_valid_files(self):
        for valid, _, _ in self.valid_files:
            self.assertEqual(_qseq_sniffer(valid), (True, {}))

    def test_qseq_sniffer_invalid_files(self):
        for invalid, _, _, _ in self.invalid_files:
            self.assertEqual(_qseq_sniffer(invalid), (False, {}))


if __name__ == '__main__':
    unittest.main()



r"""Biological Embeddings (:mod:`skbio.embedding`)
=================================================

.. currentmodule:: skbio.embedding

This module provides support for storing embeddings for biological objects, such
as protein embeddings outputted from protein language models (pLMs).


Embedding types
---------------

.. autosummary::
    :toctree: generated/

    Embedding
    SequenceEmbedding
    ProteinEmbedding


Embedding vector types
----------------------

.. autosummary::
    :toctree: generated/

    SequenceVector
    ProteinVector


Embedding vector utilities
--------------------------

.. autosummary::
    :toctree: generated/

    embed_vec_to_numpy
    embed_vec_to_distances
    embed_vec_to_ordination
    embed_vec_to_dataframe

"""

# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

from ._embedding import (
    Embedding,
    SequenceEmbedding,
    EmbeddingVector,
    SequenceVector,
    embed_vec_to_numpy,
    embed_vec_to_distances,
    embed_vec_to_ordination,
    embed_vec_to_dataframe,
)

from ._protein import ProteinEmbedding, ProteinVector, example_protein_embedding


__all__ = [
    "Embedding",
    "SequenceEmbedding",
    "EmbeddingVector",
    "SequenceVector",
    "embed_vec_to_numpy",
    "embed_vec_to_distances",
    "embed_vec_to_ordination",
    "embed_vec_to_dataframe",
    "ProteinEmbedding",
    "ProteinVector",
    "example_protein_embedding",
]



"""BLAST+7 format (:mod:`skbio.io.format.blast7`)
==============================================

.. currentmodule:: skbio.io.format.blast7

The BLAST+7 format (``blast+7``) stores the results of a BLAST [1]_ database
search. This format is produced by both BLAST+ output
format 7 and legacy BLAST output format 9. The results
are stored in a simple tabular format with headers. Values are separated by the
tab character.

An example BLAST+7-formatted file comparing two nucleotide sequences, taken
from [2]_ (tab characters represented by ``<tab>``):

.. code-block:: none

    # BLASTN 2.2.18+
    # Query: gi|1786181|gb|AE000111.1|AE000111
    # Subject: ecoli
    # Fields: query acc., subject acc., evalue, q. start, q. end, s. st\
art, s. end
    # 5 hits found
    AE000111<tab>AE000111<tab>0.0<tab>1<tab>10596<tab>1<tab>10596
    AE000111<tab>AE000174<tab>8e-30<tab>5565<tab>5671<tab>6928<tab>6821
    AE000111<tab>AE000394<tab>1e-27<tab>5587<tab>5671<tab>135<tab>219
    AE000111<tab>AE000425<tab>6e-26<tab>5587<tab>5671<tab>8552<tab>8468
    AE000111<tab>AE000171<tab>3e-24<tab>5587<tab>5671<tab>2214<tab>2130

Format Support
==============
**Has Sniffer: Yes**

+------+------+---------------------------------------------------------------+
|Reader|Writer|                          Object Class                         |
+======+======+===============================================================+
|Yes   |No    |:mod:`pandas.DataFrame`                                        |
+------+------+---------------------------------------------------------------+

Format Specification
====================
There are two BLAST+7 file formats supported by scikit-bio: BLAST+ output
format 7 (``-outfmt 7``) and legacy BLAST output format 9 (``-m 9``). Both file
formats are structurally similar, with minor differences.

Example BLAST+ output format 7 file::

    # BLASTP 2.2.31+
    # Query: query1
    # Subject: subject2
    # Fields: q. start, q. end, s. start, s. end, identical, mismatches, sbjct\
frame, query acc.ver, subject acc.ver
    # 2 hits found
    1	8	3	10	8	0	1	query1	subject2
    2	5	2	15	8	0	2	query1	subject2

.. note:: Database searches without hits may occur in BLAST+ output format 7
   files. scikit-bio ignores these "empty" records:

   .. code-block:: none

       # BLASTP 2.2.31+
       # Query: query1
       # Subject: subject1
       # 0 hits found

Example legacy BLAST output format 9 file:

.. code-block:: none

    # BLASTN 2.2.3 [May-13-2002]
    # Database: other_vertebrate
    # Query: AF178033
    # Fields:
    Query id,Subject id,% identity,alignment length,mismatches,gap openings,q.\
 start,q. end,s. start,s. end,e-value,bit score
    AF178033    EMORG:AF178033  100.00  811 0   0   1   811 1   811 0.0 1566.6
    AF178033    EMORG:AF031394  99.63   811 3   0   1   811 99  909 0.0 1542.8

.. note:: scikit-bio requires fields to be consistent within a file.

BLAST Column Types
------------------
The following column types are output by BLAST and supported by scikit-bio.
For more information on these column types, see :mod:`skbio.io.format.blast6`.

+-------------------+----------------------+
|Field Name         |DataFrame Column Name |
+===================+======================+
|query id           |qseqid                |
+-------------------+----------------------+
|query gi           |qgi                   |
+-------------------+----------------------+
|query acc.         |qacc                  |
+-------------------+----------------------+
|query acc.ver      |qaccver               |
+-------------------+----------------------+
|query length       |qlen                  |
+-------------------+----------------------+
|subject id         |sseqid                |
+-------------------+----------------------+
|subject ids        |sallseqid             |
+-------------------+----------------------+
|subject gi         |sgi                   |
+-------------------+----------------------+
|subject gis        |sallgi                |
+-------------------+----------------------+
|subject acc.       |sacc                  |
+-------------------+----------------------+
|subject acc.ver    |saccver               |
+-------------------+----------------------+
|subject accs       |sallacc               |
+-------------------+----------------------+
|subject length     |slen                  |
+-------------------+----------------------+
|q\\. start          |qstart                |
+-------------------+----------------------+
|q\\. end            |qend                  |
+-------------------+----------------------+
|s\\. start          |sstart                |
+-------------------+----------------------+
|s\\. end            |send                  |
+-------------------+----------------------+
|query seq          |qseq                  |
+-------------------+----------------------+
|subject seq        |sseq                  |
+-------------------+----------------------+
|evalue             |evalue                |
+-------------------+----------------------+
|bit score          |bitscore              |
+-------------------+----------------------+
|score              |score                 |
+-------------------+----------------------+
|alignment length   |length                |
+-------------------+----------------------+
|% identity         |pident                |
+-------------------+----------------------+
|identical          |nident                |
+-------------------+----------------------+
|mismatches         |mismatch              |
+-------------------+----------------------+
|positives          |positive              |
+-------------------+----------------------+
|gap opens          |gapopen               |
+-------------------+----------------------+
|gaps               |gaps                  |
+-------------------+----------------------+
|% positives        |ppos                  |
+-------------------+----------------------+
|query/sbjct frames |frames                |
+-------------------+----------------------+
|query frame        |qframe                |
+-------------------+----------------------+
|sbjct frame        |sframe                |
+-------------------+----------------------+
|BTOP               |btop                  |
+-------------------+----------------------+
|subject tax ids    |staxids               |
+-------------------+----------------------+
|subject sci names  |sscinames             |
+-------------------+----------------------+
|subject com names  |scomnames             |
+-------------------+----------------------+
|subject blast names|sblastnames           |
+-------------------+----------------------+
|subject super      |sskingdoms            |
|kingdoms           |                      |
+-------------------+----------------------+
|subject title      |stitle                |
+-------------------+----------------------+
|subject strand     |sstrand               |
+-------------------+----------------------+
|subject titles     |salltitles            |
+-------------------+----------------------+
|% query coverage   |qcovs                 |
|per subject        |                      |
+-------------------+----------------------+
|% query coverage   |qcovhsp               |
|per hsp            |                      |
+-------------------+----------------------+

Examples
--------
Suppose we have a BLAST+7 file:

>>> from io import StringIO
>>> import skbio.io
>>> import pandas as pd
>>> fs = '\\n'.join([
...     '# BLASTN 2.2.18+',
...     '# Query: gi|1786181|gb|AE000111.1|AE000111',
...     '# Database: ecoli',
...     '# Fields: query acc., subject acc., evalue, q. start, q. end, s. st\
art, s. end',
...     '# 5 hits found',
...     'AE000111\\tAE000111\\t0.0\\t1\\t10596\\t1\\t10596',
...     'AE000111\\tAE000174\\t8e-30\\t5565\\t5671\\t6928\\t6821',
...     'AE000111\\tAE000171\\t3e-24\\t5587\\t5671\\t2214\\t2130',
...     'AE000111\\tAE000425\\t6e-26\\t5587\\t5671\\t8552\\t8468'
... ])
>>> fh = StringIO(fs)

Read the file into a ``pd.DataFrame``:

>>> df = skbio.io.read(fh, into=pd.DataFrame)
>>> df # doctest: +NORMALIZE_WHITESPACE
       qacc      sacc        evalue  qstart     qend  sstart     send
0  AE000111  AE000111  0.000000e+00     1.0  10596.0     1.0  10596.0
1  AE000111  AE000174  8.000000e-30  5565.0   5671.0  6928.0   6821.0
2  AE000111  AE000171  3.000000e-24  5587.0   5671.0  2214.0   2130.0
3  AE000111  AE000425  6.000000e-26  5587.0   5671.0  8552.0   8468.0

Suppose we have a legacy BLAST 9 file:

>>> from io import StringIO
>>> import skbio.io
>>> import pandas as pd
>>> fs = '\\n'.join([
...     '# BLASTN 2.2.3 [May-13-2002]',
...     '# Database: other_vertebrate',
...     '# Query: AF178033',
...     '# Fields: ',
...     'Query id,Subject id,% identity,alignment length,mismatches,gap openin\
gs,q. start,q. end,s. start,s. end,e-value,bit score',
...     'AF178033\\tEMORG:AF178033\\t100.00\\t811\\t0\\t0\\t1\\t811\\t1\\t81\
1\\t0.0\\t1566.6',
...     'AF178033\\tEMORG:AF178032\\t94.57\\t811\\t44\\t0\\t1\\t811\\t1\\t81\
1\\t0.0\\t1217.7',
...     'AF178033\\tEMORG:AF178031\\t94.82\\t811\\t42\\t0\\t1\\t811\\t1\\t81\
1\\t0.0\\t1233.5'
... ])
>>> fh = StringIO(fs)

Read the file into a ``pd.DataFrame``:

>>> df = skbio.io.read(fh, into=pd.DataFrame)
>>> df[['qseqid', 'sseqid', 'pident']] # doctest: +NORMALIZE_WHITESPACE
     qseqid          sseqid  pident
0  AF178033  EMORG:AF178033  100.00
1  AF178033  EMORG:AF178032   94.57
2  AF178033  EMORG:AF178031   94.82

References
----------
.. [1] Altschul, S.F., Gish, W., Miller, W., Myers, E.W. & Lipman, D.J. (1990)
   "Basic local alignment search tool." J. Mol. Biol. 215:403-410.
.. [2] http://www.ncbi.nlm.nih.gov/books/NBK279682/


"""  # noqa: D205, D415

# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

import pandas as pd

from skbio.io import create_format, BLAST7FormatError
from skbio.io.format._blast import _parse_blast_data

blast7 = create_format("blast+7")

column_converter = {
    "query id": "qseqid",
    "query gi": "qgi",
    "query acc.": "qacc",
    "query acc.ver": "qaccver",
    "query length": "qlen",
    "subject id": "sseqid",
    "subject ids": "sallseqid",
    "subject gi": "sgi",
    "subject gis": "sallgi",
    "subject acc.": "sacc",
    "subject acc.ver": "saccver",
    "subject accs.": "sallacc",
    "subject length": "slen",
    "q. start": "qstart",
    "q. end": "qend",
    "s. start": "sstart",
    "s. end": "send",
    "query seq": "qseq",
    "subject seq": "sseq",
    "evalue": "evalue",
    "bit score": "bitscore",
    "score": "score",
    "alignment length": "length",
    "% identity": "pident",
    "identical": "nident",
    "mismatches": "mismatch",
    "positives": "positive",
    "gap opens": "gapopen",
    "gaps": "gaps",
    "% positives": "ppos",
    "query/sbjct frames": "frames",
    "query frame": "qframe",
    "sbjct frame": "sframe",
    "BTOP": "btop",
    "subject tax ids": "staxids",
    "subject sci names": "sscinames",
    "subject com names": "scomnames",
    "subject blast names": "sblastnames",
    "subject super kingdoms": "sskingdoms",
    "subject title": "stitle",
    "subject titles": "salltitles",
    "subject strand": "sstrand",
    "% query coverage per subject": "qcovs",
    "% query coverage per hsp": "qcovhsp",
    "Query id": "qseqid",
    "Subject id": "sseqid",
    "gap openings": "gapopen",
    "e-value": "evalue",
}


@blast7.sniffer()
def _blast7_sniffer(fh):
    # Smells a BLAST+7 file if the following conditions are present
    #   -First line contains "BLAST"
    #   -Second line contains "Query" or "Database"
    #   -Third line starts with "Subject" or "Query" or "Database"
    lines = [line for _, line in zip(range(3), fh)]
    if len(lines) < 3:
        return False, {}

    if not lines[0].startswith("# BLAST"):
        return False, {}
    if not (lines[1].startswith("# Query:") or lines[1].startswith("# Database:")):
        return False, {}
    if not (
        lines[2].startswith("# Subject:")
        or lines[2].startswith("# Query:")
        or lines[2].startswith("# Database:")
    ):
        return False, {}

    return True, {}


@blast7.reader(pd.DataFrame, monkey_patch=False)
def _blast7_to_data_frame(fh):
    line_num = 0
    columns = None
    skiprows = []
    for line in fh:
        if line == "# Fields: \n":
            # Identifies Legacy BLAST 9 data
            line = next(fh)
            line_num += 1
            if columns is None:
                columns = _parse_fields(line, legacy=True)
                skiprows.append(line_num)
            else:
                next_columns = _parse_fields(line, legacy=True)
                if columns != next_columns:
                    raise BLAST7FormatError(
                        "Fields %r do not equal fields %r" % (columns, next_columns)
                    )
                skiprows.append(line_num)
        elif line.startswith("# Fields: "):
            # Identifies BLAST+7 data
            if columns is None:
                columns = _parse_fields(line)
            else:
                # Affirms data types do not differ throught file
                next_columns = _parse_fields(line)
                if columns != next_columns:
                    raise BLAST7FormatError(
                        "Fields %r do not equal fields %r" % (columns, next_columns)
                    )
        line_num += 1
    if columns is None:
        # Affirms file contains BLAST data
        raise BLAST7FormatError("File contains no BLAST data.")
    fh.seek(0)

    return _parse_blast_data(
        fh,
        columns,
        BLAST7FormatError,
        "Number of fields (%r) does not equal number" " of data columns (%r).",
        comment="#",
        skiprows=skiprows,
    )


def _parse_fields(line, legacy=False):
    r"""Remove '\n' from fields line and returns fields as a list (columns)."""
    line = line.rstrip("\n")
    if legacy:
        fields = line.split(",")
    else:
        line = line.split("# Fields: ")[1]
        fields = line.split(", ")
    columns = []
    for field in fields:
        if field not in column_converter:
            raise BLAST7FormatError(
                "Unrecognized field (%r)."
                " Supported fields: %r" % (field, set(column_converter.keys()))
            )
        columns.append(column_converter[field])
    return columns
